package heasarc.util;

/** This class provides routines
  *  for efficient parsing of character arrays.
  */
public class NumberParser {

    /** Test routine 
      */
    public static void main(String[] args) {
	
	char[] a = args[0].toCharArray();
	int offset = Integer.parseInt(args[1]);
	int size = Integer.parseInt(args[2]);
	
	NumberParser n = new NumberParser(a);
	n.setOffset(offset);
	
	
	System.out.println("readDouble gives:"+n.readDouble(size));
	n.setOffset(offset);
	System.out.println("readInt gives:"+n.readInt(size));
	n.setOffset(offset);
	System.out.println("readLong gives:"+n.readLong(size));
    }
    
    /** Array being parsed */
    private char[] input;
    
    /** Current offset into character array being read. */
    private int offset;
    
    /** Length of last parsed value */
    private int numberLength;
    
    /** Did we find a sign last time we checked? */
    private boolean foundSign;
    
    /** Construct a parser. 
      * @param input	The character array to be parsed.
      */
    public NumberParser(char[] input) {
	this.input    = input;
	this.offset   = 0;
    }
    
    /** Set the offset into the array.
      * @param offset	The desired offset from the beginning
      *                  of the array.
      */
    public void setOffset(int offset) {
	this.offset = offset;
    }
    
    /** Get the current offset
     */
    public int getOffset() {
	return offset;
    }
    
    /** Get the number of characters used to parse the previous
     *  number
     */
    public int getNumberLength() {
	return numberLength;
    }

    /** Convert a segment to a double.
      * Leading spaces are ignored.
      * @param length	The maximum number of characters
      *                  used to parse this number
      */
    public double readDouble(int length) {
	
        int startOffset = offset;
	
	double number = 0;
	int i = 0;
	
	// Skip initial blanks.
	length -= skipWhite(length);
	if (length == 0) {
	    return 0;
	}
	
	double mantissaSign = checkSign();
	if (foundSign) {
	    length -= 1;
	}
	
	number = getBareInteger(length);   // This will update offset
	length -= numberLength;            // set by getBareInteger
	
	// Check for fractional values after decimal
	if (length > 0 && input[offset] == '.') {
	    
	    offset += 1;
	    length -= 1;
	    
	    double numerator = getBareInteger(length);
	    if (numerator > 0) {
		number += numerator/tenPow(numberLength);
	    }
	    length -= numberLength;
	}

	// Look for an exponent
        if (length > 0) {
	    
	    // Our Fortran heritage means that we allow 'D' for the exponent
	    // indicator.
	    if (input[offset] == 'e'  || input[offset] == 'E' ||
		input[offset] == 'd'  || input[offset] == 'D') {
		
		offset += 1;
		length -= 1;
		if (length > 0) {
		    int sign = checkSign();
		    if (foundSign) {
			length -= 1;
		    }
		    
		    int exponent = (int) getBareInteger(length);
		    number *= tenPow(exponent*sign);
		    length -= numberLength;
		}
	    }
	}
	
	numberLength = offset-startOffset;
	return mantissaSign*number;
    }
    
    
    
    /** Convert a segment to a float
      * Leading spaces are ignored.
      * @param length	The maximum number of characters
      *                  used to parse this number
      */
    public float readFloat(int length) {
	return (float)readDouble(length);
    }
    
    
    /** Convert a segment to an int
      *  Leading spaces are ignored.
      *  @param length	The maximum number of characters
      *                  used to parse this number
      */
    public int readInt(int length) {
	return (int) readLong(length);
    }
    
    
    /** Convert a segment to a long
      * Leading spaces are ignored.
      * @param length	The maximum number of characters
      *                  used to parse this number
      */
    public long readLong(int length) {
	
	int startOffset = offset;
	
	// Skip white space.
	length -= skipWhite(length);
	
	long number = 0;
	long sign = checkSign();
	if (foundSign) {
	    length -= 1;
	}
       
	number = sign * (long)getBareInteger(length);
	numberLength = offset-startOffset;
	
	return number;
    }
    
    /** Get a string (trimmed on the right)
      * @param length  The maximum length of the string.
      */
    public String readString(int length) {
	
	int i;
	for (i=length-1; i >= 0; i -= 1) { 
	    if (input[offset+i] != ' ' &&
		input[offset+i] != '\t') {
		break;
	    }
	}
	
	String result;
	if (i >= 0) {
	    result =  new String(input, offset, i+1);
	} else {
	    result = "";
	}
	offset       += length;
	numberLength = length;
	return result;
    }
    
    /** Get a boolean value.  Any string starting
      * with a 't' or 'T' is assumed to be true.
      * All other strings are false.
      * @param length The segment search length.
      */
    public boolean readBoolean(int length) {
	
	
	for (int i=0; i<length; i += 1) {
	    if (input[offset+i] == 'T' ||
		input[offset+i] == 't') {
		offset += i+1;
		numberLength = i+1;
		return true;
	    } else if (input[offset+i] != ' ' &&
		       input[offset+i] != '\t') {
		offset += i+1;
		numberLength = i+1;
		return false;
	    }
	    
	}
	numberLength = length;
	offset += length;
	return false;
    }
    
	
    /** Get the integer value starting at the current position.
      * This routine returns a double rather than an int/long
      * to enable it to read very long integers (with reduced
      * precision).  This routine does not handle leading
      * spaces.
      * 
      * @param length	The maximum number of characters to use.
      */
    private double getBareInteger(int length) {
	
	int startOffset = offset;
	double number = 0;
	
	while (length > 0 && input[offset] >= '0' && input[offset] <= '9') {
	    
	    number *= 10;
	    number += input[offset] - '0';
	    offset += 1;
	    length -= 1;
	}
	numberLength = offset-startOffset;
        return number;
    }
    
    /** Skip white space.  This routine skips with space in
      * the input and returns the number of character skipped.
      * White space is defined as ' ' or '\t'.
      * @param length The maximum number of characters to skip.
      */
    public int skipWhite(int length) {
	
	int i;
	for (i=0; i<length; i += 1) {
	    if (input[offset+i] != ' ' && input[offset+i] != '\t') {
		break;
	    }
        }
	
	offset += i;
	return i;
	
    }
    
    /** Find the sign for a number .
      * This routine looks for a sign (+/-) at the current location
      * and return +1/-1 if one is found, or +1 if not.
      * The foundSign boolean is set if a sign is found and offset is
      * incremented.
      */
    private int checkSign () {
	
	foundSign = false;
	
        if (input[offset] == '+') {
	    foundSign = true;
	    offset += 1;
	    return 1;
	} else if (input[offset] == '-') {
	    foundSign = true;
	    offset += 1;
	    return -1;
	}
	
	return 1;
    }
    
    
    /** Is a region blank?
      * This function is used when we allow a number to have
      * trailing white space.
      * @param length The length of the region to be tested
      */
    private boolean isWhite(int length) {
	
	int i;
	for (i=0; i<length; i += 1) {
	    if (input[i+offset] != ' ' && input[i+offset] != '\t') {
		break;
	    }
	}
	offset += i;
	return i == length;
    }
 
    /** Get 10^n.
      * Use this to avoid possible rounding errors in Math.pow(...).
      * It may not be as efficient.
      * @param n The power of 10 desired.
      */
    private double tenPow(int n) {
	
	double num = 1;
	for (int i=0; i < Math.abs(n); i += 1) {
	    num *= 10;
	}
	if (n < 0) {
	    num = 1/num;
	}
	return num;
    }
	
	
}
