#!/usr1/local/bin/perl5 -T

# This program extracts data product URLs from the HEASARC
# database and displays them to the requester.  This user
# must specify the table for which the data products
# are desired and values of key parameters.  The
# user may specify which data product types are desired and
# the format of the output.  Optionally the user can
# request information about just the available datasets
# and the key values.
#
# Usage:
#    getProd may be called using either a GET or POST URL
#    The HTTP keyword arguments are:
#
#    -source=		The table for which data products are requested.
#    -meta.all          Show available tables.
#    -product=		A specifically requested data product.
#    'field1' ... 	where 'field1', 'field2' are the names
#                       of the query keys for the desired table.
# Written January 29, 2001, By T. McGlynn, NASA/GSFC (USRA).

use strict;
use lib "./lib";

use Browse::Config;
Browse::Config::AlwaysRun();

my $header;

print "Content-type: text/plain\n\n";

# Get arguments.
my $in = Browse::CGI::ReadParse();

# Has the user requested a list of queryable tables?
if (defined($$in{"-meta.all"})) {
    displayTables();
    exit();
}

# Has a table been defined.
my ($table) = $$in{"-source"};

# If not then describe the command.
if (!defined($table)) {
    displayCmd(0);
    exit();
}

# See if the user has specified a specific set of products.
my $products = $$in{"-product"};
if (defined($products) ) {
    my @prod = split("\0", $products);
    $products = \@prod;
}

# Get the product sets for the table -- or just the requested ones
# if the user has made a request.

Browse::DBQ::Registry::init("databases.dat");
my $query = "select set_name,set_description,tag_format from zzdpsets where table_name='$table'";

if (defined($products)) {
    $query .= " and set_name in ('" . join("','", @$products) . "')";
}

my ($sets) = Browse::DBQ::query($query);


# Find all of the fields that are going to be substituted in.
# This includes database functions of the fields, e.g., substrings.

my %needed;
my $set;
foreach $set(@$sets) {
    my ($name, $descrip, $format) = @$set;
    my @subfields = $format =~ /(\@\{[^}]*\})/g;
    foreach (@subfields) {
        $needed{$_} = 1;
    }
}

# Now given all of these fields, find the actual database
# parameters that are needed.

my %params;
my $field;
foreach $field(keys(%needed)) {
    my @match = $field =~ /[a-zA-Z]\w*\b\s*[^(']/g;   #' match for editor
    foreach (@match) {
	chop($_);
	$params{$_} = 1;
    }
}

my @pars   = keys(%params);

# Now see if the user has actually specified
# and of the fields we need.

my %got;
foreach $field(@pars) {
    if ($$in{$field}) {
        $got{$field} = $$in{$field};
    }
}

my @ky = keys(%got);

# If the user has specified any parameters that
# actually belong to the table, then try to
# get data products.  The user need not have
# specified all parameters.  Otherwise
# describe the parameters that are needed.

if ($#ky >= 0) {
    displayProducts($table, $products, $sets, \%needed, \%got);
    
} else {
    my @need = keys(%needed);
    displayNeeded($table, $products, $sets, \@pars);
}

# Describe the required parameters for a particular table
sub displayNeeded {

    my ($table, $products, $sets, $need) = @_;

    # Indicate the input parameters.
    print "<INFO>\n Processed Parameters:\n";
    print "   -source=$table\n";
    
    if (defined($products)) {
        foreach (@$products) {
            print "    -product=$_\n";
	}
    }
    
    print <<EOT;
</INFO>
<RESOURCE id="main">
  <TITLE> Required fields for table $table </TITLE>
  <DESCRIPTION> 
     This resource describes the fields required
     to determine the HEASARC data products associated with
     table $table.
  </DESCRIPTION>
EOT


    if (!defined($products)) {
        print <<EOT;
   <FIELD id='product' name='-product' value='A' width='30'>
      <DESCRIPTION> Limit the request to the specified product.
                    This keyword may be repeated.
      </DESCRIPTION>
      <VALUES>
EOT
        my $set;
	foreach $set(@$sets) {
	    my ($name, $descrip, $format) = @$set;
	    print <<EOT;
         <OPTION name='$name'> $descrip </OPTION>
EOT
        }
        print <<EOT;
     </VALUES>
   </FIELD>
EOT
    }
      
    # Get information about the required parameters.
    
    $query = "select parameter_name, parameter_format, ".
         "         parameter_minval, parameter_maxval from zzpar "   .
         "   where table_name='$table' and parameter_name in('"  .
	 join("','", @$need) . "')";
	 
    my ($params) = Browse::DBQ::query($query);

    my $par; 
    my $specfld = "-source=$table"; 
    if (defined($products)) {
	foreach (@$products) {
	    $specfld .= "&-product=".Browse::CGI::Encode($_);
	} 
    }

    foreach $par(@$params) {
        my ($name, $form, $min, $max) = @$par;
	my $width;
	
	# Convert the format field to the Astrores format.
	if ($form =~ /char/) {
	
	    $form =~ s/char/A/i;
	    $width = " width='".substr($form,1)."'";
	    $form = substr($form,0,1);
	    
	} else {
	    $form = uc(substr($form, 0, 1));
        }
        print <<EOT;
    <FIELD name='$name' datatype='$form' $width>
	<VALUES>
	   <MIN value='$min' />
	   <MAX value='$max' />
	</VALUES>
    </FIELD>
EOT
    }
    
    print <<EOT;
  <LINK content-role='query'
     action='/cgi-bin/tam/w3x/getProd.pl?$specfld&' />";
</RESOURCE>
EOT
}

# Describe this command.
sub displayCmd {

    my ($meta) = @_;
    
    if ($meta) {
        print <<EOT;
<INFO>
    Processed Parameters: -meta.all
</INFO>
EOT
    }

    print<<EOT;
<RESOURCE ID="main"> 
  <TITLE>HEASARC Data Products<QUERY>
  <DESCRIPTION>
  This interface permits the user to determine the
  data products available within the HEASARC archive
  associated with any catalog in the HEASARC database.
  </DESCRIPTION>

  <FIELD id="table" name="-source" width="20" datatype="A">
    <DESCRIPTION>
    This table for which associated data products are requested.
    </DESCRIPTION>
EOT
    if (!$meta) {
        print <<EOT;
    <LINK content-role="hints" href="/cgi-bin/getProd.pl?-meta.all" />
EOT
    }
    
    print <<EOT;
  </FIELD>
  <FIELD id="product" name="-product" width=100 datatype="A">
    <DESCRIPTION>
        A specifically requested product.  This field may be repeated.
    </DESCRIPTION>
  </FIELD>
  <LINK content-role="query" content-type="text/plain"
        action="/cgi-bin/getProd.pl?" />
EOT

    if (!$meta) {
        print "</RESOURCE>\n";
    }
}

# Describe the requested data products.
sub displayProducts {

    my ($table, $products, $sets, $need, $got) = @_;
    
    displaySetInfo($table, $products, $sets, $got);

    # Now process all of the sets.
    processSets($table, $sets, $need, $got);
}


# Process sets for selected observation.
sub processSets {

    my ($table, $sets, $need, $got) = @_;

    my %translations = {};
    
    # Get the parameter formats for the selected parameters.
    # (We need to know whether to enclose parameter values with 
    #  apostrophes)
    
    my $parfield = "'" . join("','", keys(%$got)) . "'";
    
    my $query = "select parameter_name, parameter_format from zzpar ".
                           " where table_name = '$table' and ".
			   "   parameter_name in( $parfield )";
    
    my ($res) = Browse::DBQ::query($query);
    my %forms;
    my $row;
    foreach $row (@$res) {
        $forms{$$row[0]} = $$row[1];
    }
    
    # For each substitution field in the data products
    # get the translation for the specified data values.
    
    my $key;
    foreach $key (keys(%$need)) {
	my $trans = getTranslation($key, $got, \%forms);
	$translations{$key} = $trans;
    }
	
print <<EOT;
  <DATA><CSV headlines="2" colsep="|">
  <![CDATA[
set|format|product|type
-|-|-|-|-
EOT

    my $set;
    foreach $set(@$sets) {
	    
        my($name, $descrip, $format) = @$set;
	
	foreach $key(keys(%$need)) {
	
	    # Put in the translations into the set description.
	    my $trans = $translations{$key};
	    my $len = length($key);
	    while (index($format, $key) >= 0) {
	        substr($format, index($format, $key), $len) = $trans;
	    }
	}
	# Now split into individual tag fields.
        my @tags = split(',', $format);

	my $tag;
	foreach $tag(@tags) {
	    processTag($name, $descrip, $tag);
	}
    }
    
    print <<EOT;
]></CSV>
  </DATA>
</TABLE>
</RESOURCE>
EOT
}

# Print a summary of selected data products sets (i.e., the set description).
# This will only be printed once at the top of the output rather than
# repeated for all selected observations.

sub displaySetInfo {

    my ($table, $products, $sets, $got) = @_;
    print <<EOT;
<INFO>
    Processed parameters:
        -source=$table
EOT

    if ($products) {
        foreach (@$products) {
            print "        -product=$_\n";
	}
    }
    my $field;
    foreach $field(keys(%$got)) {
        print "        $field=$$got{$field}\n";
    }
    
    print <<EOT;
</INFO>
<RESOURCE id="main"> 
  <TITLE>HEASARC Archive Products</TITLE>

<TABLE>
  <NAME>       $table </NAME>
  <TITLE>      Data Products associated with $table </TITLE>
  <DESCRIPTION>
        Files and directories associated with data products
	that are keyed to the $table given the specified keys.
  </DESCRIPTION>
  
  <FIELD name="set" datatype='A' width=20>
    <DESCRIPTION> Data set </DESCRIPTION>
    <VALUES>
EOT
    foreach $set(@$sets) {
	print <<EOT;
      <OPTION value="$$set[0]"> $$set[1] </OPTION>
EOT
    }
	
	
    print <<EOT;
  </FIELD>
  <FIELD name='format' datatype='A' width='15'>
    <DESCRIPTION> The format of the described file </DESCRIPTION>
    <VALUES type='legal'>
      <OPTION value='image/x-fits'> A FITS file -- possibly compressed </OPTION>
      <OPTION value='image/gif'>  A GIF image </OPTION>
      <OPTION value='text/postscript'> Postscript text or plots </OPTION>
      <OPTION value='text/plain'> ASCII text </OPTION>
      <OPTION value='directory'> A set of datafiles in a directory </OPTION>
    </VALUES>
  </FIELD>
  <FIELD name="product" datatype="A" width='100'>
     <DESCRIPTION> A link to the data product </DESCRIPTION>
     <LINK content-role="query" content-type="\$\{format\}"/ 
       href="ftp://legacy.gsfc.nasa.gov/\$\{product\}" />
  </FIELD>
  <FIELD name="type" datatype="A" width='40'>
    <DESCRIPTION> Type of the product </DESCRIPTION>
  </FIELD>
EOT
}

# Process a single tag string.  This
# may contain a wild card and thus refer to multiple tages.
sub processTag {

    my ($name, $descrip, $tag) = @_;
    
    $tag =~ tr/A-Z/a-z/;
    if ($tag =~ /\*/) {
        $tag =~ s/\*/%/;
	$query = "select dp_format, dp_type, dp_url from zzdp where dp_tag like '$tag'";
    } else {
	$query = "select dp_format, dp_type, dp_url from zzdp where dp_tag = '$tag'";
    }
    
    my ($res) = Browse::DBQ::query($query);
	
    my $line;
    foreach $line(@$res) {
	my ($format, $type, $url) = @$line;
	$url = Browse::DPL::resolveShortcuts($url);
	displayTag($name, $format, $type, $url);
    }
}

# Display a single tag.
sub displayTag {

    my %types = (
       "FITS" => "image/x-fits",
       "GIF" =>  "image/gif",
       "POSTSCRIPT" => "text/postscript",
       "ASCII" => "text/plain",
       "DIRECTORY" => "directory"
    );
    
    my ($set, $fileFormat, $type, $url)  = @_;

    # Convert the HEASARC format type to a mime type.
    my $cont = $types{$fileFormat}; 
    
    # Chop off the header.
    $url =~ s/ftp\:\/\/legacy\.gsfc\.nasa\.gov\///; 
    
    print "$set|$cont|$url|$type\n";
}

# Conver the @{...} fields in a data set format into
# the values appropriate for the current row.
sub getTranslation {

    my ($string, $got, $formats) = @_;
    
    $string =~ s/^\@\{//;
    $string =~ s/\}$//;
    my $fld;
    
    foreach $fld(keys (%$got)) {
        if ($fld eq $string) {
	    return $$got{$fld};
	}
    }
    
    my @arr = keys(%$got);
    
    sub lenSrt {
         length($b) <=> length($a);
    }
    
    my @sarr = sort lenSrt @arr;
    
    foreach $fld (@sarr) {
        my $quote;
	my $form = $$formats{$fld};
	if ($form =~ /char/i) {
	    $quote = "'";
	}
	my $sub = $quote.$$got{$fld}.$quote;
	$string =~ s/$fld/$sub/g;
    }
    
    my $query = "select $string";
    my ($res) = Browse::DBQ::query($query);
    return $$res[0][0];
}
	    
# Process an error.
sub error {
    print "<ERROR_IN_STREAM>\n",@_,"\n</ERROR_IN_STREAM>\n";
    exit();
}
    

sub displayTables {

    my $sql = "select a.table_name,a.table_description,b.parameter_value ".
              "   from zzgen a, zzext b ".
	      "   where a.table_name=b.table_name and ".
	      "   a.table_name in(select distinct table_name from zzdpsets) and ".
	      "   b.parameter_name='observatory_name'";
	      
    $sql .= " order by b.parameter_value";
    
    
    Browse::DBQ::Registry::init("databases.dat");
    my ($results) = Browse::DBQ::query($sql);
    
    displayCmd(1);
    
    
    print <<EOT;

 <!-- Output fields follow -->
 <TABLE ID="heasarc_table_list">
  <FIELD name="table" width="20" format="A">
     <DESCRIPTION>
         A HEASARC table name
     </DESCRIPTION>
  </FIELD
  <FIELD name="description" width=60 format="A">
      <DESCRIPTION> A one line description of the table.
      </DESCRIPTION>
  <FIELD name="mission" width="40" format="A">
     <DESCRIPTION>
       The mission with which the table is associated.
     </DESCRIPTION>
  </FIELD>
  <DATA><CSV headlines= colsep="|">
  <<![CDATA[
table | description | mission
--- | --- | ---
EOT
    my $line;
    foreach $line(@$results) {
        print join("|", @$line),"\n";
    }

    print "]></CSV></DATA>\n </TABLE>\n</RESOURCE>\n";
}
