# $Id: CrossRequest.pm,v 1.22 2005/03/07 18:34:20 lmm Exp $
# $Log: CrossRequest.pm,v $
# Revision 1.22  2005/03/07 18:34:20  lmm
# Added log entry to SQL log.
#
# Revision 1.21  2003/07/31 18:47:40  cwh
# Added a new setPositions, so that users can reissue Cross queries with new
#   positions.
#
# Revision 1.20  2003/04/28 02:24:15  cwh
# Fixed bug  - was setting format for offset to int, not float.
#
# Revision 1.19  2002/10/17 19:18:57  cwh
# just a little cleaning up.
#
# Revision 1.18  2002/09/05 18:22:05  cwh
# changed the way the dpsets are handled.
#
# Revision 1.17  2002/03/13 20:52:57  cwh
# Tom fixed this:
#   Steve reported an error that when one did an anticorrelation
#   starting from a query result, the anticorrelation returned
#   results that weren't in the original result.  Basically the
#   where clause of the original result wasn't being propogated properly.
#
# Revision 1.16  2001/12/18 15:42:35  cwh
# Don't undef the sort-related object properties.
#
# Revision 1.15  2001/12/11 19:49:01  cwh
# Don't show sort arrows for cross products of Vizier tables yet.
#
# Revision 1.14  2001/12/06 16:47:24  cwh
# Call setOffsetVar() and getOffsetVar() to isolate the offset var name.
# Set $$self{sortable} = up|down.
# Call makeSort().
#
# Revision 1.13  2001/11/30 19:56:13  cwh
# Removed distance() sub, now calls distance() in the Utils module.
#
# Revision 1.12  2001/11/07 23:34:17  cwh
# Moved everything except DBD/ into lib/Browse/, so as not to pollute the module
# namespace with our Sort, Search, CGI, etc.  This required putting Browse:: in
# front of every module call and package declaration.  Also any mod::new() had
# to be changed to mod->new().
#
# Revision 1.11  2001/10/25 19:26:23  cwh
# Changed _Offset to _Search_Offset for clarity.
#
# Revision 1.10  2001/06/20 22:57:32  cwh
# Got rid of $::writeHeader in calls to ::Error.
#
# Revision 1.9  2001/06/15 03:23:58  tam
# Add table_name to tableInfo structure.
#
# Revision 1.8  2001/06/14 15:39:21  tam
# Fixed bugs in Anticorrelation queries.
#    - Sybase syntax for allow_dup_row
#    - Eliminate prefix from key field
#
# Revision 1.7  2001/06/08 19:46:37  cwh
# Now passing $::writeHeader to ::Error calls, so we don't get repeating headers.
#
# Revision 1.6  2001/05/22 17:52:03  cwh
# Added Id and Log to files without them.
#

use strict;

use Browse::QueryRequest;
use Browse::CachedQuery;

# This package defines a Request object that can
# be used to cross-correlate two tables.  If both
# of the underlying tables are local, then the query
# is generated directly.  Otherwise a CachedQuery
# object is created for the remote table.

package Browse::CrossRequest;

my($DTOR) = 0.017453292519943295769236907684886127134428718885417;
# arcminutes to radians constant (Pi/(180*60))
my($MTOR) = $DTOR/60;

use vars qw( @ISA );
@ISA = qw( Browse::HeasarcQuery );

# Generate a query given two earlier queries and
# parameters which describe the cross-correlation.

sub new {

    my ($self, $r1, $r2, $parameters) = @_;
    
    my $obj = {};
    
    if (!$r1->localQuery()) {
        $r1 = Browse::CachedQuery->new($r1);
    }
    if (!$r2->localQuery()) {
        $r2 = Browse::CachedQuery->new($r2);
    }
    
    $$obj{r1} = $r1;
    $$obj{r2} = $r2;
    
    $$obj{parameters} = $parameters;
    
    bless $obj;
    
    return $obj;
}

# Get the next prefix.
sub nextPrefix {

    my ($self) = @_;
    
    my $r2 = $$self{r2};
    if ($r2) {
        return $r2->nextPrefix();
    } else {
        return undef;
    }
}

# Set the prefix for the first table used in this query.
sub setPrefix {

    my ($self, $prefix) = @_;
    
    if (!$prefix) {
        $prefix = 'a';
    }
    
    my $r1 = $$self{r1};
    my $r2 = $$self{r2};
    $r1->setPrefix($prefix);
    $r2->setPrefix($r1->nextPrefix());
    $$self{prefix} = $prefix;
}
    
## Postions around which to do a cone search.
sub setPositions {
    my ($self, $positions) = @_;
    # Get the composing queries.
    my $r1 = $$self{r1};
    my $r2 = $$self{r2};
    $$r1{"positions"} = $positions if defined $$r1{"positions"};
    $$r2{"positions"} = $positions if defined $$r2{"positions"};
}
	
# Get the metadata associated with the cross-correlation results.    
sub getMeta {

    my ($self) = @_;
    
    my $prefix = $self->getPrefix();
    if (!$prefix) {
        $prefix = "a";
    }
    $$self{r1}->setPrefix($prefix);
    $$self{r2}->setPrefix($$self{r1}->nextPrefix());
    $self->setOffsetVar("_xoffset");
    
    
    my $params = $$self{"parameters"};
    
    # Get the composing queries.
    my $r1 = $$self{r1};
    my $r2 = $$self{r2};
    
    # Initialize the composing queries.
    $r1->makeQuery();
    $r2->makeQuery();
    
    # Check again if we have to cache this
    # result before using it.  E.g., something
    # like the USNO catalog or a remote DBI catalog.
   
    if (!$r1->localQuery()) {
        $r1 = Browse::CachedQuery->new($r1);
	$$self{r1} = $r1;
    }
    
    if (!$r2->localQuery()) {
        $r2 = Browse::CachedQuery->new($r2);
	$$self{r2} = $r2;
    }
    
    my $m1 = $r1->putMeta();
    my $m2 = $r2->putMeta();
    
    
    $$self{table} = "$$m1{table} x $$m2{table}";
    
    
    my $t1 = $$m1{tableinfo};
    my $t2 = $$m2{tableinfo};
 
    my $type = "Cross ";
    my $prep = "with";
    if ($$params{"anticorrelation"}) {
        $type = "Anti";
	$prep = "against";
    }
    
    
    my %tinfo;
    $tinfo{"table_name"} = $$self{"table"};
    $tinfo{"xradius"} = $$params{"xradius"} if $$params{"pcross"};
    $tinfo{"xtime"} = $$params{"xtime"} if $$params{"tcross"};
    $tinfo{"table_description"} = $type."Correlation of $$t1{table_description} $prep ".
                                  "$$t2{table_description}";
    ($tinfo{"table_location"})  = Browse::DBQ::Registry::getDefault();
    
    $tinfo{"default_search_radius"} = $$t1{"default_search_radius"};
    $tinfo{"table_document_url"}    = $$t1{"table_document_url"}.'|'.$$t2{"table_document_url"};
    
    $tinfo{"right_ascension"} = $$t1{"right_ascension"}."|".$$t2{"right_ascension"};
    $tinfo{"declination"}     = $$t1{"declination"}.    "|".$$t2{"declination"};
    $tinfo{"target_name"}     = $$t1{"target_name"}.    "|".$$t2{"target_name"};
    $tinfo{"start_time"}      = $$t1{"start_time"}.     "|".$$t2{"start_time"};
    
   
    $$self{tableinfo} = \%tinfo;
    my %pinfo;
    
    my $p1 = $$m1{paraminfo};
    my $p2 = $$m2{paraminfo};
    
    foreach (keys(%$p1)) {
        $pinfo{$_} = $$p1{$_};
	
    }
    
    foreach (keys(%$p2)) {
        $pinfo{$_} = $$p2{$_};
    }
    
    $$self{paraminfo} = \%pinfo;
    
    my $f1 = $$r1{fields};
    replaceCone($r1, $f1);
    
    if ($$params{"anticorrelation"}) {
        $$params{"secondary"} = 0;
    }

    my $secondary = $$params{"secondary"};
    if ($secondary) {
        my $f2 = $$r2{fields};
	replaceCone($r2, $f2);
    
        # Put the index fields for table 2 at the
        # beginning.
    
        my $sk1 = 0;
        foreach (@$f1) {
            my $x = $_;
            my ($a,$b) = split(" as ", $x);
	    if ($b) {
	        $x = $b;
	    }
            if ($x =~ /^w3\_/i) {
	        $sk1 += 1;
	    } else {
	        last;
	    }
	}
    
        my $sk2;
        foreach (@$f2) {
            my $x = $_;
            my ($a,$b) = split(" as ", $x);
	    if ($b) {
	        $x = $b;
	    }
            if ($x =~ /^w3\_/i) {
	        $sk2 += 1;
	    } else {
	        last;
	    }
        }
        # Extract the key fields from the second table.
        my @flds = splice(@$f2, 0, $sk2);
    
        # Insert them after the key fields from the first table.
        splice(@$f1, $sk1, 0, @flds);
    
        $$self{fields}    = [@$f1, @$f2];
    } else {
        my @flds = @$f1;
	$$self{fields}  = \@flds;
    }
    
    my $fld = $$self{fields};
    
    
    if ($$params{"pcross"}) {
        my $xd = Browse::Utils::distance(chip($$t1{"right_ascension"}),
	                  chip($$t1{"declination"}),
			  chip($$t2{"right_ascension"}),
			  chip($$t2{"declination"}));
	if ($secondary) {
            my $name;
            if ($self->getPrefix()) {
                $name ="_".$self->getPrefix().$self->getOffsetVar;
            } else {
                $name = $self->getOffsetVar;
            }
	    my $nf = "$xd/$MTOR as $name";

	    push (@$fld, $nf);
	    my $paraminfo = $$self{"paraminfo"};
	    my $hash = {
	        "table_name"       => $$self{"table"},
	        "parameter_name"   => $name,
		"parameter_format" => "float4:.3f",
		"parameter_unit"   => "'"
	    };
	    $$paraminfo{$name} = $hash;
	}
	$$self{cosoffset} = $xd;
    }
    
    if ($$params{"tcross"}) {
        my $xd = "abs(".chip($$t1{start_time})."-".chip($$t2{start_time}).")";
        my $name;
        if ($self->getPrefix()) {
            $name ="_".$self->getPrefix()."_tdelta"
        } else {
            $name = "_tdelta";
        }
	if ($secondary) {
	    push(@$fld, "$xd as _tdelta");
	    my $paraminfo = $$self{"paraminfo"};
	    my $hash = {
	        "table_name"       => $$self{"table"},
	        "parameter_name"   => $name,
		"parameter_format" => "float4:.3f",
		"parameter_unit"   => "days"
	    };
	    $$paraminfo{$name} = $hash;
	}
	$$self{timedelta} = $xd;
    }
	
    
    # Need to mark data products with the table they are
    # associated with.  However if we're already dealing
    # with a cross-correlation, then this has already been done.
    
    my $dpsets;
    
    my $tab = $$t1{table_name};
    my $x = $$m1{dpsets};
    foreach (keys %$x) {
        if ($tab !~ / x /){
            $$dpsets{"$_ $tab"}=$$x{$_};
	} else {
            $$dpsets{"$_"}=$$x{$_};
	}
    }
    if ($secondary) {
        my $tab = $$t2{table_name};
        $x = $$m2{dpsets};
    
        foreach (keys %$x) {
            if ($tab !~ / x /){
                $$dpsets{"$_ $tab"}=$$x{$_};
	    } else {
                $$dpsets{"$_"}=$$x{$_};
	    }
        }
    }
        
    
    $$self{dpsets}    = $dpsets;
    
    
    $$self{keyinfo} = $$m1{keyinfo};
    if ($secondary){
        $$self{keyinfo} .= '||'.$$m2{keyinfo};
    }
    
    #$$self{positions} = undef;
    #$$self{sorts}     = undef;
    #$$self{dirs}      = undef;
    #$$self{quals}     = undef;
    $$self{sortable}  = undef;
    $$self{sortable}  = "up|down" unless $$t1{table_name} =~ /\//; #no Viz!
    
    return {
    
        table      => $$self{table},
	tableinfo  => $$self{tableinfo},
	paraminfo  => $$self{paraminfo},
	dpsets     => $$self{dpsets},
	keyinfo    => $$self{keyinfo},
	fields     => $$self{fields},
	positions  => $$self{positions},
	sorts      => $$self{sorts},
	dirs       => $$self{dirs},
	quals      => $$self{quals},
	sortable   => $$self{sortable},
	serialization => $$self{serialization},
	type       => "meta"
	};
}

# Get the actual expression needed to generate an offset field.
sub replaceCone {

    my ($r1, $o1) = @_;
    
    my $p = $r1->getPrefix();
    foreach (@$o1) {
        if ($_ eq "_$p".$r1->getOffsetVar()) {
	    $_ = $$r1{"coneParam"};
	}
    }
}
    
# Get the next piece of data from a request.
sub getResult {

    my ($self) = @_;
    
    my $status = $$self{status};
    my $params = $$self{parameters};
    
    if (!$status || $status eq "initiate") {
        $$self{status} = "data";
	$$self{nrows} = 0;
	return $self->getMeta();
    }
    
    my $nrow = $$self{nrows};
    
    my $r1 = $$self{r1};
    my $r2 = $$self{r2};
    $self->makeSort();
    
    my $sql;
    
    if ($nrow == 0) {
    
        $sql = "select ";
	if (!$$params{"secondary"}) {
	    $sql .= "distinct ";
	}
	if ($$params{"anticorrelation"}) {
	    $sql .= $self->getAntifields();
	} else {
	    $sql .= $self->getFields();
	}
	$sql .= " from "; 
	$sql .= $self->getSource();
	$sql .= $self->getWhere();
	$sql .= $self->getOrder();

	
	$$self{"sql"} = $sql;
print "xcorr sql=$sql<br>\n" if $::debug;
	
        my $conn = $$self{"tableinfo"}{"table_location"};
        my ($db) = Browse::DBQ::Registry::getConnection($conn);
        $$self{"DBQ"} = $db;
	if ($$params{"anticorrelation"}) {
	    $self->runAnti();
	}
        Browse::Log::SqlLogger($$self{"sql"},$conn);
	
        $self->startQuery();
	if ($DBI::errstr) {
	    Browse::Utils::Error("Error starting cross-correlation query: $DBI::errstr");
	}
    }
    
    
    my $line = $self->getDataLine();
    if (!defined($line)) {
        $self->qclose();
    } else {
        $$self{nrows} += 1;
    }
    
    return $line;
}

# Get the fields associated with the query
sub getFields {

    my ($self) = @_;
    
    
    my $flds = $$self{fields};
    
    my $params = $$self{"parameters"};
    
    if ($$params{"minimal"}) {
        my $tableinfo = $$self{"tableinfo"};
	
	my $needed = $$tableinfo{"right_ascension"};
	$needed   .=  "|". $$tableinfo{"declination"};
        $needed   .=  "|". $$tableinfo{"start_time"};
	$needed   .=  "|". $$tableinfo{"target_name"};
	
	my @xflds = split('\|', $needed);
	map $_ =~ s/^@//, @xflds;
	
	push(@xflds, ("_tdelta", $self->getOffsetVar()));
	
	my %hashNeed;
	map $hashNeed{$_} = 1, @xflds;
	
	my @uflds;
	
        my $fld;
	my $cnt = 0;
	foreach $fld(@$flds) {
	
	    my $f = $fld;
	    my ($xx,$yy) = split(" as ", $fld);
	    if ($yy) {
	        $f = $yy;
	    }
	    if (substr($f, 0, 3) =~ /w3_/i) {
	        push(@uflds, $fld);
	    } elsif ($cnt == 0) {
	        push(@uflds, $fld);
		$cnt += 1;
	    } elsif ($hashNeed{$f}) {
	        push(@uflds, $fld);
	    }
	}
	$flds = \@uflds;
    }
    return join(",", @$flds);
}

# Get the source tables of the query
sub getSource {

    my ($self) = @_;
    
    my $r1 = $$self{r1};
    my $r2 = $$self{r2};
    
    return $r1->getSource() . "," . $r2->getSource();
}


# Get the where clause associated with the query
sub getWhere {

    my ($self) = @_;
    
    my $r1 = $$self{r1};
    my $r2 = $$self{r2};
    
    my $params = $$self{parameters};
	
    my $sql   = " where ";
    
    my $wh;
    my $and;
    
    my $req;
    
    foreach $req (($r1, $r2)) {
        $wh = $req->getWhere();
	
        if ($wh) {
            $wh =~ s/^ *where +//;
	    $sql .= $and.$wh;
	    $and = " and ";
	}
    }
	
    if ($$params{pcross}) {
        my $dec1 = chip($$r1{tableinfo}{"declination"});
	my $dec2 = chip($$r2{tableinfo}{"declination"});
	my $rad = $$params{"xradius"}/60;
	$sql .= "$and ($dec1 between $dec2-$rad and  $dec2+$rad) and ";
	
	$sql .= "(($$self{cosoffset}) < ".($rad*$DTOR).") "; 
	$and = " and ";
    }
	
    if ($$params{tcross}) {
	$sql.= "$and ($$self{timedelta}) < $$params{xtime} ";
        $and = " and ";
    }
    return $sql;
}

# This converts a parameter from the form a_name to a.name
sub chip {

    my ($str) = @_;
    
    $str =~ s/^\@//;
    $str =~ s/^(.)\_/$1\./;
    return $str;
    
}

# Get the fields needed for the initial anticorrelaiton query
sub getAntifields {

    my ($self) = @_;
    
    my $params = $$self{parameters};
    my $r1     = $$self{"r1"};
    my $t1     = $$r1{"tableinfo"};
    
    my @antis;
    
    if ($$params{"pcross"}) {
   	push(@antis, chip($$t1{"declination"}));
        push(@antis, chip($$t1{"right_ascension"}));
    }
    if ($$params{"tcross"}) {
        push(@antis, chip($$t1{"start_time"}));
    }
    my $flds = join(",", @antis) . " into #ax$$ ";
    
    $$self{antifields} = \@antis;
    $$self{antitable}  = "#ax$$";
    
    return $flds;
    
}

# Run little correlation query and then ready
# the main query.
sub runAnti {

    my ($self) = @_;
    
    # This creates a table with the correlating values.
    Browse::DBQ::execute($$self{"sql"});
    
    if ($DBI::errstr) {
        Browse::Utils::Error("Error creating temporary table for anticorrelation. $DBI::errstr");
    }
    
    my $table = $$self{"antitable"};
    my $flds  = $$self{"antifields"};
    my $tx    = $table."x";
    my $f0    = $$flds[0];
    my $r1    = $$self{"r1"};
    $f0 =~ s/[^ ]\.//;
    
    my $sql = "create clustered index $tx on $table ($f0) with allow_dup_row";
    Browse::DBQ::execute($sql);
    if ($DBI::errstr) {
        Browse::Utils::Error("Error indexing temporary table for anticorrelation. $DBI::errstr");
    }
    
    $sql =   "select ";
    $sql .=  $self->getFields();
    
    my $prefix = $r1->getPrefix();
    my $where = " where ";
    my $wh = $r1->getWhere();
    $sql .=  " from ".$r1->getTable();
    if ($wh) {
        $sql .= $wh . " and ";
	$where = "";
    }	
    
    $sql .= " $where not exists(select * from $table where ";
    my $fld;
    my $and;
    foreach $fld (@$flds) {
        my $f = chip($fld);
	$sql .= "$and$f = ";
	substr($f, 0, 1) = $table;
	$sql .= $f;
	$and = " and ";
    }
    $sql .= ')';
    $$self{sql} = $sql;
    
}


1;	
      
