use strict;

package Browse::XMLDisplay;


# Display the results in Astrores style XML.
# The current version simply strips out the key columns.

use vars qw( @ISA );
@ISA = qw( Browse::TextDisplay );
use Browse::TextDisplay;


# Indicate that we're going to have an Astrores XML output.
sub pageHeader {
    my ($self) = @_;
    
    $self->setCoordFormat("decimal");
    $self->setDateFormat("mjd");
    
    $$self{table_count} = 0;
    
    print "Content-type: text/plain\n" .
	"Content-Disposition: attachment; filename=browse_results.xml\n\n";

    print <<EOT;
<!DOCTYPE ASTRO SYSTEM "http://vizier.u-strasbg.fr/xml/astrores.dtd">
<ASTRO ID="v0.9" xmlns:ASTRO="http://vizier.u-strasbg.fr/doc/astrores.htx">
  <DESCRIPTION about="http://heasarc.gsfc.nasa.gov">
     HEASARC Browse data service
     Please send inquiries to mailto:request\@athena.gsfc.nasa.gov
  </DESCRIPTION>
EOT
		  
}

# Print the table header up till the parameter list.
# We defer that until we know the order of output parameters.
sub tableHeader {

    my ($self) = @_;
    
    $$self{tableopen} = 1;
    
    if ($$self{tableCount} == 0) {
        my $system;
	my $equinox;
    
        # Handle the definitions -- these should only be printed once
	# for the entire request.
	
        if (lc($$self{coordsys}) eq "equatorial") {
	    if ($$self{equinox} <= 1975) {
	        $system="eq_FK4";
	    } else {
	        $system="eq_FK5";
	    }
	    $equinox = $$self{equinox}
	    
	} elsif (lc($$self{coordsys}) eq "galactic") {
	    $system="galactic";
	    
	}
	if ($system)  {
	
	    my $cstring = "system=\"$system\"";
	    if ($equinox) {
	        $cstring .= " equinox=\"$equinox\"";
	    }
	    
	    print qq(
  <DEFINITIONS>
    <COOSYS $cstring />
  </DEFINITIONS>
);

        }
    }
    
    $$self{tableCount} += 1;
    
    my $resid = 9000+$$self{tableCount};
    my $meta  = $$self{"meta"};
    my $tab   = $$meta{"tableinfo"}; 
    my $nam   = $$tab{"table_name"};
    my $desc  = $$tab{"table_description"};
    my $url   = $$tab{"table_document_url"};
    
    # Since Browse resources are exactly one table the names and titles
    # of the resource are redundant with the table.
    
    print <<EOT;
    
  <RESOURCE ID="$resid">
    <NAME> $nam </NAME>
    <TITLE> $desc </TITLE>
    <TABLE ID="$nam\_$resid">
      <NAME> $nam </NAME>
      <TITLE> $desc </TITLE>
EOT

    $$self{rowCount} = 0;

}

sub legend {}

# Terminate the 'page' by closing the
# ASTRO element.
sub pageFooter {
    print "</ASTRO>\n";
}

# Process each line of data.  When we process the
# first line we can finish writing out the data header.

sub processData {

    my ($self, $data) = @_;

    if (!$$self{"transform"}) {
        $$self{"transform"} = $self->setTransform($data);
    }

    # Print the field descriptions -- can't do
    # this earlier since we don't have the order until now.
    
    if ($$self{rowCount} == 0) {
    
	my $meta  = $$self{"meta"};
	my $names = $$data{"names"};
	my $types = $$data{"types"};
	
	my $paraminfo = $$meta{"paraminfo"};
	
	my $nskip = 0;
	while ($nskip <= $#$names) {
	    if (substr($$names[$nskip], 0, 3) !~ /w3_/i) {
	        last;
	    }
	    $nskip += 1;
	}
	
	$$self{"skipFields"} = $nskip;
        
	# Get rid of the unique key columns.
	splice(@$names, 0, $nskip);
	splice(@$types, 0, $nskip);
	
	# Loop over each column
        my $i;	
	for ($i=0; $i <= $#$names; $i += 1) {
	
	    my $name = $$names[$i];
	    my $type = $$types[$i];
	    my $unit = $$paraminfo{$name}{"parameter_unit"};
	    my $desc = $$paraminfo{$name}{"parameter_description"};
	    my $form;
	    
	    $unit =~ s/^ *//;
	    $unit =~ s/ *$//;
	    if (length($unit) <= 0) {
	        $unit = undef;
	    } else {
	        $unit = "unit=\"$unit\"";
	    }
	    
	    
	    if ($type <= 1  || $name eq "class") {
	        $form=  "A";
	    } elsif ($type eq 6) {
	          $form=  "D"
	    } elsif ($type eq 7) {
	        $form = "F";
	    } else {
	        $form = "I";
	    }

            if ($name eq "_Search_Offset") {
	       $form=  "A";
            } 
	    
	    
	    print qq(
	    
      <FIELD name="$name" datatype="$form" $unit >
        <DESCRIPTION> $desc </DESCRIPTION>
      </FIELD>
);
        }
	
	print qq(
      <DATA>
        <CSV colsep="|" headlines="2"><![CDATA[
);
        print join("|", @$names), "\n\n";
    }
    
    
    # Print the data.
    
    $$self{rowCount} += 1;
   
    my $trans = $$self{"transform"};
    $data = $$data{"data"};
    Browse::TableFormat::transformLine($data, $trans);

    my $nskip = $$self{skipFields};
    splice(@$data, 0, $nskip);
    print join("|", @$data),"\n";
    
}


sub closeTable {
    my ($self) = @_;
    if ($$self{tableopen}) {
        print "]]></CSV></DATA>\n";
        print "     </TABLE>\n";
        print "   </RESOURCE>\n";
	delete($$self{tableopen});
    }
}

sub YourSearchWas {
}
sub resultsSummary {}
sub showEmpties {}
sub beforeNewTable {}
sub afterTable {}

## Successful load
1;
