#=============================================================================#
# Module Name:    Astro::Config
# Module Author:  Edward J. Sabol <sabol@lheamail.gsfc.nasa.gov>
# Creation Date:  1998/06
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Miscellaneous configuration variables used by the Astro package.
#=============================================================================#
#
# $Id: Config.pm,v 1.8 2004/03/16 04:13:09 sabol Exp $
# $Log: Config.pm,v $
# Revision 1.8  2004/03/16 04:13:09  sabol
# Added $ambiguous_match_is_error variable for configuring whether or not Astro
#   returns an error when the name resolver returns multiple, ambiguous
#   results.
#
# Revision 1.7  2003/02/12 19:32:26  sabol
# Fixed typo in default setting of $converter.
#
# Revision 1.6  2003/02/03 21:02:21  sabol
# Updated comments style. Now uses Exporter to optionally export the
#   configuration variables to any client namespace.
#   Astro::Utils::convert_using_SLA() is now the default for $converter.
#   Added the following configuration variables: $ambiguous_match_tolerance,
#   $resolver_cache_directory, $SIMBAD_user, $SIMBAD_password,
#   $SIMBAD_executable, and $NED_executable.
#
# Revision 1.5  2001/05/10 22:06:20  sabol
# Imported into HEASARC CVS archive.
#
# Revision 1.4  2000/08/31 15:02:15  sabol
# Added $decimal_accuracy package configuration variable.
#
# Revision 1.3  1999/08/11 11:42:22  sabol
# Now uses 'vars' compiler directive to specify package-wide variables.
#   Moved all 'use' directives after the package declaration.
#
# Revision 1.2  1999/08/10 18:36:15  sabol
# Coco configuration variable now falls back to /heasarc/bin/coco automatically
#   if /www/htdocs/cgi-bin/W3Browse/bin/coco either doesn't exist or isn't
#   executable for whatever reason.
#
# Revision 1.1  1998/06/16 13:33:34  sabol
# Created.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::Config;

use strict;
use Astro::Utils;
use Astro::Utils::SLAConvert;

require Exporter;
use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION
	    $converter $cleaner $error_handler $coco $eol $decimal_accuracy
	    $ambiguous_match_tolerance $ambiguous_match_is_error
	    $resolver_cache_directory
	    $SIMBAD_user $SIMBAD_password $SIMBAD_executable $NED_executable);
$VERSION = '1.01';
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw($converter $cleaner $error_handler $coco $eol $decimal_accuracy
		$ambiguous_match_tolerance $resolver_cache_directory
		$SIMBAD_user $SIMBAD_password $SIMBAD_executable
		$NED_executable);

#------------------------------------------------------------------------------

$converter = \&Astro::Utils::SLAConvert::convert_using_SLA;
	# reference to subroutine for converting coordinates

$cleaner = \&Astro::Utils::clean;
	# reference to subroutine for "cleaning" up Astro::Object input
	#   values; by default, it is assigned a reference to
	#   &Astro::Utils::clean

$error_handler = defined(&CGI::Error) ? \&CGI::Error
  : defined(&::Error) ? \&::Error : undef;
	# reference to subroutine for handling (usually fatal constructor
	#   errors) in Astro::Object objects; by default, it is assigned a
	#   reference to either &CGI::Error or &::Error() or undef

$coco = (-x '/www/htdocs/cgi-bin/W3Browse/bin/coco')
  ? '/www/htdocs/cgi-bin/W3Browse/bin/coco' : '/heasarc/bin/coco';
	# name (including full path) of the external executable application
	#   'coco' used for converting Astro::Object's from one coordinate
	#   system to another

$eol = '<br>';
	# string to use as end-of-line marker in debugging output; it
	#   defaults to "<br>", but you should set it to "\n" at runtime unless
        #   the debugging output is intended for a Web browser

$decimal_accuracy = 5;
	# the default number of digits to keep after the decimal point when
	#   formatting decimal degrees

$ambiguous_match_tolerance = 0.167;
	# the default tolerance for considering singular matches from multiple
	#   name resolvers to be unambiguously the same; value in arcminutes
	#   (default is 0.167 arcminutes or approximately 10 arcseconds)

$ambiguous_match_is_error = 1;
	# boolean indicates whether Astro should report an error if an object
	#   name is resolved ambiguously

$resolver_cache_directory = '/www/htdocs/cgi-bin/W3Browse/dbm';
	# location (complete path) to the directory containing the name
	#   resolver caches; if undef, then the default value specified in
	#   Astro::Object::Resolver is used instead

$SIMBAD_user = 'u616idsv';
	# user ID for SIMBAD name resolver service (used by
	#   Astro::Object::Resolver::SIMBAD)

$SIMBAD_password = 'yjqjmzex';
	# password for SIMBAD name resolver service (used by
	#   Astro::Object::Resolver::SIMBAD)

$SIMBAD_executable =
  -x '/www/htdocs/cgi-bin/W3Browse/bin/multisnr'
  ? '/www/htdocs/cgi-bin/W3Browse/bin/multisnr'
  : -x '/www/htdocs/cgi-bin/W3Browse/bin/snr'
    ? '/www/htdocs/cgi-bin/W3Browse/bin/snr'
    : -x '/dbase/lib/simbad/multisnr'
      ? '/dbase/lib/simbad/multisnr'
      : '/dbase/lib/simbad/snr';
	# location (complete path) of the SIMBAD name resolver executable
	#   (used by Astro::Object::Resolver::SIMBAD)

$NED_executable =
  -x '/www/htdocs/cgi-bin/W3Browse/bin/multiNED'
  ? '/www/htdocs/cgi-bin/W3Browse/bin/multiNED'
  : -x '/www/htdocs/cgi-bin/W3Browse/bin/NED_name'
    ? '/www/htdocs/cgi-bin/W3Browse/bin/NED_name'
    : -x '/dbase/lib/simbad/multiNED'
      ? '/dbase/lib/simbad/multiNED'
      : '/dbase/lib/simbad/NED_name';
	# location (complete path) of the NED name resolver executable
	#   (used by Astro::Object::Resolver::NED)

#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
