#=============================================================================#
# Module Name:    Astro::CoordSys
# Module Author:  Edward J. Sabol
# Creation Date:  1997/05
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Object-oriented implementation of astronomical coordinate systems.
#=============================================================================#
#
# $Id: CoordSys.pm,v 1.10 2003/02/03 21:04:47 sabol Exp $
# $Log: CoordSys.pm,v $
# Revision 1.10  2003/02/03 21:04:47  sabol
# Updated comments style.
#
# Revision 1.9  2001/05/11 20:25:50  sabol
# Quiet bogus compiler warning about usage of undefined $equinox_prefix.
#
# Revision 1.8  2001/05/02 18:30:36  sabol
# Ecliptic coordinates don't require an equinox, but they should specify an
#   epoch in MJD. The epoch date is optional, however. (Note that SLA defaults
#   to 2000/1/1 for the epoch if it is not specified.) The epoch MJD attribute
#   is stored as a sub_type attribute and can be accessed via the sub_type()
#   method. Added epoch test to equals() method.
#
# Revision 1.7  2001/05/02 14:22:25  sabol
# A "B" or "J" prefix on the equinox is now interpreted as indicating either
#   FK4 or FK5 Equatorial.
#
# Revision 1.6  1999/08/11 11:45:55  sabol
# Moved all 'use' directives after the package declaration.
#
# Revision 1.5  1999/08/10 18:37:21  sabol
# Improved method documentation and added object class overview.
#
# Revision 1.4  1998/11/25 13:55:26  sabol
# The equinox can now optionally have a decimal point and digits after the
#   decimal point.
#
# Revision 1.3  1998/06/30 16:31:24  sabol
# Fixed error handler argument list. Increased compiler efficiency by
#   changing double-quoted constant strings to single-quoted strings.
#
# Revision 1.2  1998/06/16 15:55:22  sabol
# Now uses Astro::Config. Replaced all references to Astro::Utils with
#   Astro::Config.
#
# Revision 1.1  1998/06/09 18:35:25  sabol
# Updated comments.
#
# Revision 1.0  1997/05/01 15:28:04  sabol
# Initial release.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::CoordSys;

use strict;
use Astro::Config;

#------------------------------------------------------------------------------
# Object class overview:
#
# See documentation for individual methods for more details.
#
# Public methods:
#	new			Creates a new object.
#	initialize		Initializes an existing object.
#	equals			Compares two coordinate system objects.
#	equinox			Returns the coordinate system's equinox.
#	name			Returns the name of the coordinate system.
#	short			Returns the short name of the coordinate system.
#	sub_type		Returns the sub-type of the coordinate system.
#
# Private methods:
#	None!
#
# Also, do not access internal data members!
#
#------------------------------------------------------------------------------
sub new {
    # Creates blessed reference to astronomical coordinate system object.
    #
    # Constructor arguments:
    #   anonymous hash with the following keys and corresponding values
    #     system  : one of {"Equatorial","Galactic","Ecliptic"};
    #                 comparison is case-insensitive and partial matches
    #                 are allowed); defaults to "Equatorial"
    #     equinox : decimal number, usually one of {1950,2000,2050};
    #                 optional if system is "Galactic"; otherwise, default: 2000
    #     epoch_mjd : decimal MJD, optional, only used if system is "Ecliptic"
    #
    # Note: An undefined reference is returned if a constructor error is
    #       encountered.

    my($this) = shift;
    my($class) = ref($this) || $this;
    my($self) = {};
    bless $self, $class;
    $self = $self->initialize(@_);
    return($self);

} #eosub--Astro::CoordSys::new
#------------------------------------------------------------------------------
sub initialize {
    # Arguments: Same as new().

    my($self) = shift(@_);
    my(%args) = @_;
    my($system) = (defined($args{'system'}) ? $args{'system'} : 'Equatorial');
    my($equinox) = (defined($args{'equinox'}) ? $args{'equinox'} : 2000);
    my($epoch_mjd) = (defined($args{'epoch_mjd'}) ? $args{'epoch_mjd'} : '');
    my($error_handler) = ((defined($Astro::Config::error_handler)
			   && ref($Astro::Config::error_handler))
			  ? $Astro::Config::error_handler : sub { });
    my($error) = '';

    if (($system !~ /(Galactic|Ecliptic)/i)
	&& ($equinox !~ /^[BJ]?\d+(\.\d*)?$/)) {
	$error = 'Error: Invalid equinox.';
	&$error_handler($error);
	$self = undef;
	return($self);
    }
    if (($system =~ /Ecliptic/i)
	&& ($epoch_mjd ne '') && ($epoch_mjd !~ /^\d+(\.\d*)?$/)) {
	$error = 'Error: Invalid epoch MJD.';
	&$error_handler($error);
	$self = undef;
	return($self);
    }

    my($equinox_prefix) = '';
    if ($equinox =~ /^([BJ])/) {
	$equinox_prefix = $1;
	$equinox =~ s/^[BJ]//;
    }

    if ($equinox =~ /^\d+\.$/) { chop($equinox); }

    if ($system =~ /Galactic/i) {
	$self->{'name'}    = 'Galactic';
	$self->{'equinox'} = '';
	$self->{'short'}   = 'G';
	$self->{'coco'}    = 'G';
    } elsif ($system =~ /Ecliptic/i) {
	$self->{'name'}    = 'Ecliptic';
	$self->{'equinox'} = '';
	$self->{'epoch_mjd'} = $self->{'sub_type'} = $epoch_mjd;
	$self->{'short'}   = 'Ec';
	$self->{'coco'}    = 'E';
    } elsif ($system =~ /Equatorial/i) {
	$self->{'name'}    = 'Equatorial';
	$self->{'equinox'} = $equinox;
	$self->{'short'}   = 'Eq';
	if (($equinox_prefix eq 'J') || ($system =~ /(Julian|FK5)/i)) {
	    $self->{'coco'} = '5';
	    $self->{'sub_type'} = 'FK5';
	} elsif (($equinox_prefix eq 'B') || ($system =~ /(Besselian|FK4)/i)) {
	    $self->{'coco'} = '4';
	    $self->{'sub_type'} = 'FK4';
	} else {
	    $self->{'coco'} = (($equinox < 2000) ? '4' : '5');
	    $self->{'sub_type'} = (($equinox < 2000) ? 'FK4' : 'FK5');
	}
    } else {
	$error = 'Error: Invalid coordinate system specified.';
	&$error_handler($error);
	$self = undef;
    }

    return($self);
    
} #eosub--Astro::CoordSys::initialize
#------------------------------------------------------------------------------
sub equals {
    # Compares two Astro::CoordSys objects and returns non-zero value
    # if they are equivalent and zero otherwise.
    #
    # Method argument:
    #   an Astro::CoordSys object

    my($self,$other) = @_;

    return(0) if ((scalar(@_) != 2) || (ref($self) ne ref($other)));
    return((($self->{'short'} eq $other->{'short'})
	    && ($self->{'equinox'} == $other->{'equinox'})
	    && ((!defined($self->{'epoch_mjd'})
		 && !defined($other->{'epoch_mjd'}))
		|| ($self->{'epoch_mjd'} == $other->{'epoch_mjd'}))) ? 1 : 0);
    
} #eosub--Astro::CoordSys::equals
#------------------------------------------------------------------------------
sub equinox {
    # Returns the equinox value for the coordinate system.
    #
    # Method arguments: none

    my($self) = shift(@_);

    return($self->{'equinox'});
    
} #eosub--Astro::CoordSys::equinox
#------------------------------------------------------------------------------
sub name {
    # Returns the name of the coordinate system.
    #
    # Method arguments: none

    my($self) = shift(@_);

    return($self->{'name'});
    
} #eosub--Astro::CoordSys::name
#------------------------------------------------------------------------------
sub sub_type {
    # Returns the coordinate system sub-type.
    #
    # Method arguments: none

    my($self) = shift(@_);

    return($self->{'sub_type'});
    
} #eosub--Astro::CoordSys::sub_type
#------------------------------------------------------------------------------
sub short {
    # Returns the short name or abbreviation for the coordinate system.
    #
    # Method arguments: none

    my($self) = shift(@_);

    return($self->{'short'});
    
} #eosub--Astro::CoordSys::short
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
