#=============================================================================#
# Module Name:    Astro::Object::Resolver
# Module Author:  Edward J. Sabol <sabol@lheamail.gsfc.nasa.gov>
# Creation Date:  1997/05
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Object-oriented interface to an astronomical name-to-coordinates
#   resolver service, like NED or SIMBAD. This is a virtual class.
#   The actual name resolving should be done by an object class than
#   inherits from this class. Such an object should override the
#   initialize() class method and supply a do_resolve() class method
#   that implements the name resolving specifically.
#=============================================================================#
#
# $Id: Resolver.pm,v 1.11 2003/02/12 19:31:06 sabol Exp $
# $Log: Resolver.pm,v $
# Revision 1.11  2003/02/12 19:31:06  sabol
# Switched from antiquated dbmopen()/dbmclose() to tie()/untie() and
#   GDBM_File, mostly for Linux compatibility. Also, the 'use_cache' option
#   now only controls whether the cache is looked into and not whether the
#   cache should be updated. This effectively makes the 'use_cache' and
#   'cache_read_only' options orthogonal.
#
# Revision 1.10  2003/02/03 21:37:26  sabol
# Updated comments style. Added support for resolvers returning multiple
#   matches. Duplicate resolvers in the resolver list specification are now
#   removed. CheckCaches now automatically checks the caches in the same order
#   as listed in the resolver specification.
#
# Revision 1.9  2002/09/05 20:11:09  sabol
# If there is an instantiation error in new(), return undef after calling the
#   error handler. Also, the error handler is now called only if it is defined.
#
# Revision 1.8  2001/12/04 18:24:46  sabol
# Fixed modes used in dbmopen() calls. Report add_to_cache() errors if debug
#   mode is turned on.
#
# Revision 1.7  2001/05/11 20:44:56  sabol
# If you create a resolver object with multiple types and one of the types is
#   CheckCaches, then any resolver that comes after CheckCaches in the list of
#   resolver types will not use the cache for that resolver.
#
# Revision 1.6  2001/03/28 17:37:18  sabol
# Changed add_to_cache() method so that it will not add object names with
#   wildcards to the cache.
#
# Revision 1.5  2001/03/28 16:56:41  sabol
# Defer checking of illegal characters in names to the name resolver objects.
#
# Revision 1.4  1999/08/11 11:49:19  sabol
# Moved all 'use' directives after the package declaration. Entries retrieved
#   from a cache are now detainted.
#
# Revision 1.3  1998/06/17 18:42:18  sabol
# Added support for Astro::Object::Resolver::CheckCaches subclass. Also, the
#   'type' argument in the initialize() method will create any arbitrary object
#   as long as it inherits from Astro::Object::Resolver.
#
# Revision 1.2  1998/06/16 14:52:40  sabol
# Now uses Astro::Config instead of Astro::Utils. Replaced all references to
#   Astro::Utils with Astro::Config. Replaced all "<br>" and "\n" strings with
#   $Astro::Config::eol.
#
# Revision 1.1  1998/06/09 17:54:40  sabol
# The resolve() method now also returns the cache name if the name resolver
#   cache was used.
#
# Revision 1.0  1997/05/01 15:28:04  sabol
# Initial release.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::Object::Resolver;

use strict;
use GDBM_File;
use Astro::Config;
use Astro::Object::Resolver::NED;
use Astro::Object::Resolver::SIMBAD;
use Astro::Object::Resolver::CheckCaches;

#------------------------------------------------------------------------------
# Location of cache directory. Symbol table alias to Astro::Config for
# backwards compatibility with older code.
*Astro::Object::Resolver::default_cache_directory =
  \$Astro::Config::resolver_cache_directory;

#------------------------------------------------------------------------------
# Object class overview:
#
# See documentation for individual methods for more details.
#
# Public methods:
#	new			Creates a new object.
#	resolve			Resolves the object name into coordinates.
#	name			Returns the name of the resolver.
#	next			Returns the next name resolver object to try.
#	check_cache		Returns coordinates from cache for name.
#	add_to_cache		Adds a name/coordinates pair to the cache.
#
# Private methods:
#	do_resolve (virtual)	Do not use this method. Internal use only.
#                               Object classes which inherit from this class
#				should define this method.
#	initialize		Do not use this method. Internal use only.
#                               Object classes which inherit from this class
#				should redefine this method.
#
# Also, do not access internal data members. Consider them opaque.
#
#------------------------------------------------------------------------------
sub new {
    # Creates blessed reference(s) to astronomical name resolver(s).
    # Basically, it's a factory method.
    #
    # Constructor arguments:
    #   anonymous hash with the following keys and corresponding values
    #     type            : string containing name of the name resolver to
    #                         instantiate
    #     use_cache       : zero means don't look in the cache; default: 1
    #     cache_read_only : non-zero means don't write to the cache; default: 0
    # check_all_resolvers : non-zero means that all resolvers in the
    #                         resolver linked list should be checked; default: 0
    #
    # Note: An undefined reference is returned if a constructor error is
    #       encountered.

    my($this) = shift;
    my($class) = ref($this) || $this;
    my($self) = {};
    bless $self, $class;
    my %args = @_;
    $self = $self->initialize(@_);
    if ($args{'check_all_resolvers'}) {
	$self->{'check_all_resolvers'} = 1;
    }
    return($self);

} #eosub--Astro::Object::Resolver::new
#------------------------------------------------------------------------------
sub initialize {
    # Method arguments: Same as new().

    my($self) = shift(@_);
    my(%args) = @_;

    if (defined($args{'type'})) {
	my($type) = $args{'type'};
	delete($args{'type'});
	my %uniquify = ();
	my @resolvers = grep(!$uniquify{$_}++,split(/\s*\/\s*/,$type));
	my($resolver,$resolver_ref,$resolver_list_head,$resolver_list_tail);
	$resolver_list_head = $resolver_list_tail = undef;
	foreach $resolver (@resolvers) {
	    $resolver_ref = undef;
	    if (uc($resolver) eq 'SIMBAD') {	# User wants SIMBAD...
		$resolver_ref = Astro::Object::Resolver::SIMBAD->new(%args);
	    } elsif (uc($resolver) eq 'NED') {	# User wants NED...
		$resolver_ref = Astro::Object::Resolver::NED->new(%args);
	    } elsif ($resolver eq 'CheckCaches') { # User wants CheckCaches...
		$args{'use_cache'} = 0;
		my %checkcaches_args = %args;
		if (scalar(@resolvers) > 1) {
		    $checkcaches_args{'caches'} = [ grep($_ ne 'CheckCaches',
							 @resolvers) ];
		}
		$resolver_ref = Astro::Object::Resolver::CheckCaches->new(%checkcaches_args);
	    } elsif (grep($_ eq 'Astro::Object::Resolver',
			  eval "\@\{${resolver}::ISA\}")) {
		# User has his own resolver object that inherits from
		# Astro::Object::Resolver...
		$resolver_ref = $resolver->new(%args);
	    } else {
		&$Astro::Config::error_handler('Error: Specfied name resolver type either does not exist or does not inherit from Astro::Object::Resolver virtual base class.') if (ref($Astro::Config::error_handler));
		return(undef);
	    }
	    if (ref($resolver_ref)) {
		if (!ref($resolver_list_head)) {
		    $resolver_list_head = $resolver_list_tail = $resolver_ref;
		} else {
		    $resolver_list_tail->{'next'} = $resolver_ref;
		    $resolver_list_tail = $resolver_ref;
		}
	    }
	}
	$self = $resolver_list_head;
    } else {
	$self = undef;
    }

    return($self);

} #eosub--Astro::Object::Resolver::initialize
#------------------------------------------------------------------------------
sub name {
    # Returns the name resolver's name.
    #
    # Method arguments: none

    my($self) = shift(@_);

    return($self->{'name'});
    
} #eosub--Astro::Object::Resolver::name
#------------------------------------------------------------------------------
sub next {
    # Returns the next node in the name resolver linked list or undef if none.
    #
    # Method arguments: none

    my($self) = shift(@_);

    return($self->{'next'});

} #eosub--Astro::Object::Resolver::next
#------------------------------------------------------------------------------
sub resolve {
    # Provides generic support for local caches and implements the basics
    # of name resolving.
    #
    # Method arguments: Astro::Object object with a non-zero-length 'name'

    my($self) = shift(@_);
    my($object) = shift(@_);
    my($name) = $object->name();

    my($output_coord_sys) = $object->coordsys();
    my $all_matches = undef;
    my $multiple_matches = 0;
    my($coords) = ($self->{'use_cache'}
		   ? $self->check_cache($name) : 'Warning: Not using cache.');

    my($cache_used) = '';
    if ($coords =~ /(error|warning)/i) {	# Object not found in the cache.

	($coords,$all_matches) = $self->do_resolve($name);
	return($coords) if ($coords =~ /error/i);
	$cache_used = $self->{'cache_used_to_resolve'}
	  if $self->{'cache_used_to_resolve'};
	$multiple_matches = ((ref($all_matches) eq 'ARRAY')
			     && (scalar(@$all_matches) > 1)) ? 1 : 0;
	unless ($cache_used || $self->{'cache_read_only'}
		|| $multiple_matches) {
	    my($error) = $self->add_to_cache('name' => $name,
					     'coordinates' => $coords);
	    print $error,$Astro::Config::eol if $::debug;
	}
    } elsif ($self->{'use_cache'}) {
	$cache_used = $self->{'cache'};
    }
    my($coord1,$coord2) = split(/,/,$coords,2);
    $coord1 = &$Astro::Config::cleaner($coord1);
    $coord2 = &$Astro::Config::cleaner($coord2);
    unless ($output_coord_sys->equals($self->{'coordsys'})) {
	my($new_coord1,$new_coord2) =
	  &$Astro::Config::converter($coord1,$coord2,$self->{'coordsys'},
				     $output_coord_sys);
	return('Error: A problem was encountered in converting the ' .
	       'coordinates of the resolved object name to the desired ' .
	       'coordinate system.') if ($new_coord1 =~ /error/i);

	($coord1,$coord2) = ($new_coord1,$new_coord2);
    }

    if (defined($all_matches) && scalar(@$all_matches)) {
	return($coord1,$coord2,$cache_used,$all_matches);
    } else {
	return($coord1,$coord2,$cache_used);
    }

} #eosub--Astro::Object::Resolver::resolve
#------------------------------------------------------------------------------
sub check_cache {
    # Checks to see if supplied name is in the name resolver's cache.
    #
    # Method arguments: a string containing the name to be looked up.
    # Return value: a string containing a pair of coordinates (separated by
    #   a comma) or a string describing any error encountered.

    my($self,$name) = @_;
    my($error) = 'Error: Object name not found in name resolver cache.';
    my($dbm) = $self->{'cache'};

    return($error) unless ($name);
    return('Warning: Name resolver does not have a cache.') unless ($dbm);

    $name =~ tr/a-z/A-Z/;
    print 'Astro::Object::Resolver::check_cache: name: ',$name,
          $Astro::Config::eol if $::debug;

    my(%DBM);
    eval {
	tie %DBM, 'GDBM_File', $dbm, GDBM_READER, 0666;
    };
    return('Error: Could not open $dbm cache for reading.') if $@;

#### Used for testing package for case of uncached names...
##    $::debug = 1;
##    &Util::DumpHash(\%DBM) if $::debug;
##    delete $DBM{$name} if $::debug;
##    &Util::DumpHash(\%DBM) if $::debug;

    my($coords) = $DBM{$name};

    untie %DBM;

    if (! $coords) {	# Name is not in the cache...
	$coords = $error;
	print 'Astro::Object::Resolver::check_cache: no entry found for ',
	      $name,$Astro::Config::eol if $::debug;
    } else {
	print 'Astro::Object::Resolver::check_cache: found coordinates ',
	      $coords,$Astro::Config::eol if $::debug;

	my($RAd,$Decd) = split(/,/,$coords);
	$RAd  = &$Astro::Config::cleaner($RAd);
	$Decd = &$Astro::Config::cleaner($Decd);
	($RAd  =~ /^([-0-9\.\+]+)$/) && ($RAd  = $1);
	($Decd =~ /^([-0-9\.\+]+)$/) && ($Decd = $1);
	$coords = "$RAd, $Decd";
    }

    return($coords);

} #eosub--Astro::Object::Resolver::check_cache
#------------------------------------------------------------------------------
sub add_to_cache {
    # Adds a name and its corresponding pair of coordinates to the cache.
    #
    # Method arguments: a hash with the following keys:
    #   'name'        => name of the object that you want added to the cache
    #   'coordinates' => string containing a pair of coordinates (separated
    #                      by a comma)
    # Return value: a string describing any error encountered, zero-length
    #   if no error

    my($self) = shift(@_);
    my(%args) = @_;
    my($dbm) = $self->{'cache'};

    return('Warning: Name resolver does not have a cache.') unless ($dbm);
    return('Error: Improper format for adding a source to the object ' .
	   'name cache. The object name must have a non-zero length, ' .
	   'and the coordinates must be comma-delimited.')
      if ((! $args{'name'}) || ($args{'coordinates'} !~ /,/));
    return('Error: Object names with wildcards should not be cached.')
      if ($args{'name'} =~ /[\*\?]/);

    # Clean coordinatess before adding to cache...
    my($coord1,$coord2) = split(/\s*,\s*/,$args{'coordinates'});
    $coord1 = &$Astro::Config::cleaner($coord1);
    $coord2 = &$Astro::Config::cleaner($coord2);

    if ($::debug) {
	print 'Astro::Object::Resolver::add_to_cache: (coord1,coord2) = ',
	      "($coord1,$coord2)",$Astro::Config::eol;
	print 'Astro::Object::Resolver::add_to_cache: coord1 not decimal ',
	      "($coord1)",$Astro::Config::eol
		if ($coord1 !~ /^(\+|-)?[0-9]+\.[0-9]+$/);
	print 'Astro::Object::Resolver::add_to_cache: coord2 not decimal ',
	      "($coord2)",$Astro::Config::eol
		if ($coord2 !~ /^(\+|-)?[0-9]+\.[0-9]+$/);
    }

    return('Error: Improper format for adding a source to the object ' .
	   'name cache. Coordinates must be in decimal degrees.')
      if (($coord1 !~ /^[-+]?[0-9]+\.[0-9]+$/)
	  || ($coord2 !~ /^[-+]?[0-9]+\.[0-9]+$/));

    ($coord1 =~ /^([-+]?[0-9]+\.[0-9]+)$/) && ($coord1 = $1);
    ($coord2 =~ /^([-+]?[0-9]+\.[0-9]+)$/) && ($coord2 = $1);
    $args{'coordinates'} = "$coord1, $coord2";

    print "Astro::Object::Resolver::add_to_cache: opening $dbm...",
          $Astro::Config::eol if $::debug;

    # Clean object name before adding to cache...
    $args{'name'} =~ tr/a-z/A-Z/;	# Uppercase.
    $args{'name'} = &$Astro::Config::cleaner($args{'name'});

    my(%DBM);
    eval {
	tie %DBM, 'GDBM_File', $dbm, GDBM_WRCREAT, 0666;
    };
    return('Error: Could not open $dbm cache for updating.') if $@;

    if ($::debug) {
	print "Astro::Object::Resolver::add_to_cache: opened $dbm successfully",
	       $Astro::Config::eol;
	print "Astro::Object::Resolver::add_to_cache: adding $args{'name'} ",
              ":==: ($args{'coordinates'})",$Astro::Config::eol;
	print "\$DBM{$args{'name'}} = $args{'coordinates'}",
              $Astro::Config::eol;
    }

    $DBM{$args{'name'}} = $args{'coordinates'};

#    &Util::DumpHash(\%DBM) if $::debug;

    untie %DBM;

    print "Astro::Object::Resolver::add_to_cache: $args{'name'} added to ",
          $dbm,$Astro::Config::eol if $::debug;

    return('');

} #eosub--Astro::Object::Resolver::add_to_cache
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
