#=============================================================================#
# Module Name:    SLA
# Module Authors: Edward J. Sabol, Thomas McGlynn
# Creation Date:  1999/09
# Project:        HEASARC
#=============================================================================#
#
# $Id: SLA.pm,v 1.6 2003/02/03 20:54:50 sabol Exp $
# $Log: SLA.pm,v $
# Revision 1.6  2003/02/03 20:54:50  sabol
# Updated comments style. No changes to the code.
#
# Revision 1.5  2001/05/10 21:44:48  sabol
# Imported into HEASARC CVS archive.
#
# Revision 1.4  2001/05/03 15:02:15  sabol
# Added slaEpb() and slaEpj2d() for completeness.
#
# Revision 1.3  1999/08/26 14:29:51  sabol
# Added dsign() and slaDrange() functions. Modified the slaEqecl(), slaEcleq(),
#   slaEqgal(), slaGaleq(), slaPm(), slaPreces(), and slaFk45z() functions to
#   use slaDranrm() and/or slaDrange() to match the return values of the
#   SLALIB C routines exactly. Some comments from the SLALIB C source code
#   were restored to the Perl source code for consistency.
#
# Revision 1.2  1999/08/25 18:24:28  sabol
# Added Perl versions of the SLALIB C routines slaAddet(), slaSubet(),
#   slaEtrms(), slaDranrm(), slaDvdv(), dmod().
#
# Revision 1.1  1999/08/12 11:18:47  sabol
# Now uses Exporter to export the most commonly used functions to the main
#   namespace and to allow the export of all the other functions. Changed 51544
#   date in slaEcleq() and slaEqecl() to 51544.5 since the J2000.0 standard
#   epoch corresponds to 1 January 2000 12:00 UT, as defined by the
#   _Astronomical Almanac_. Ecliptic conversions now match the results you
#   get with `coco' exactly. Also, cleaned up the whitespace and comments.
#
# Revision 1.0  1996  Tom McGlynn
# Derived Perl versions of select SLALIB C routines.
#
#=============================================================================#

package SLA;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(slaEqgal slaGaleq slaEqecl slaEcleq
             slaFk45z slaFk54z slaFk524 slaPreces
             slaAddet slaSubet slaPm);
@EXPORT_OK = qw(slaPrebn slaPrec slaEcmat slaEpj slaEpb slaEpj2d slaEpb2d
                slaDcs2c slaDcc2s slaDmxv slaDimxv slaDeuler
                dmod dsign slaDranrm slaDrange slaDvdv slaEtrms 
                $DAS2R $DPI $D2PI @eqgal_rmat);
$VERSION = '1.04';

#  Derived from C SLALib routines. The C source for the SLALIB routines should
#  be consulted to understand the range of validity of these results (e.g.,
#  with regard to the precession calculations).

#  Included routines:

#  Basic conversion utilities:
#  slaEqgal             Convert from J2000 to Galactic
#  slaGaleq             Convert from Galactic to J2000
#  slaEqecl             Convert from J2000 to Ecliptic 2000
#  slaEcleq             Convert from Ecliptic 2000 to J2000
#  slaFk45z             Simple conversion from B1950 to J2000
#  slaFk54z             Simple conversion from J2000 to B1950
#  slaFk524             Full conversion from J2000 to B1950
#  slaPreces            Precess Equatorial coordinates

#  Transformation matrices:
#  slaPrebn             Calculate Besselian precession matrix
#  slaPrec              Calculate Eulerian precession matrix
#  slaEcmat             Calculate transformation matrix from Equatorial
#                           to Ecliptic

#  The following two time conversions are approximate and to be used
#  only for the purposes of calculating tranformation matrices:
#  slaEpj               Calculate Julian epoch (year) corresponding to MJD
#  slaEpb               Calculate Besselian epoch (year) corresponding to MJD
#  slaEpj2d             Calculate MJD from Julian epoch (year)
#  slaEpb2d             Calculate MJD from Besselian epoch (year)

#  Math/Astro utilities:
#  slaDcs2c             Convert coordinates to unit vector
#  slaDcc2s             Convert unit vector to coordinates
#  slaDmxv              Multiply vector by matrix
#  slaDimxv             Multiply vector by inverse matrix
#  slaPm                Add proper motion to a coordinate
#  slaEtrms             Compute the e-terms vector
#  slaAddet             Add the e-terms to a pair of FK4 coordinates
#  slaSubet             Subtract the e-terms from a pair of FK4 coordinates
#  slaDeuler            Calculate transformation matrix given Euler angles
#  slaDrange            Normalize angle into range +/- pi
#  slaDranrm            Normalize angle into range 0-2 pi
#  slaDvdv              Scalar product of two 3-vectors
#  dsign(A,B)           Magnitude of A with sign of B
#  dmod                 Modulo function


#  Transformation matrix from J2000 to Galactic
@eqgal_rmat = (-0.054875539726,
               -0.873437108010,
               -0.483834985808,
                0.494109453312,
               -0.444829589425,
                0.746982251810,
               -0.867666135858,
               -0.198076386122,
                0.455983795705);

#  pi/(180*3600):  arcseconds to radians
$DAS2R = 4.8481368110953599358991410235794797595635330237270e-6;
#  pi
$DPI = 3.1415926535897932384626433832795028841971693993751;
#  2pi
$D2PI = 6.2831853071795864769252867665590057683943387987502;

#------------------------------------------------------------------------------
#  Transformation from J2000.0 equatorial coordinates to
#  IAU 1958 Galactic coordinates.
sub slaEqgal {

   my ($dr, $dd) = @_;
   my ($dl, $db); # return values

   my (@v1, @v2);

#  Spherical to Cartesian
   @v1 = slaDcs2c ( $dr, $dd );

#  Equatorial to Galactic
   @v2 = slaDmxv ( \@eqgal_rmat, \@v1 );

#  Cartesian to spherical
   ( $dl, $db ) = slaDcc2s ( @v2 );

#  Express in conventional ranges
   $dl = slaDranrm ( $dl );
   $db = slaDrange ( $db );

   return ( $dl, $db );

} #eosub--slaEqgal
#------------------------------------------------------------------------------
#  Transformation from IAU 1958 Galactic coordinates to
#  J2000.0 equatorial coordinates.
sub slaGaleq {

   my ($dl, $db) = @_;
   my ($dr, $dd); # return values

   my (@v1, @v2);

#  Spherical to Cartesian
   @v1 = slaDcs2c ( $dl, $db );

#  Galactic to equatorial
   @v2 = slaDimxv ( \@eqgal_rmat, \@v1 );

#  Cartesian to spherical
   ( $dr, $dd ) = slaDcc2s ( @v2 );

#  Express in conventional ranges
   $dr = slaDranrm ( $dr );
   $dd = slaDrange ( $dd );

   return ( $dr, $dd );

} #eosub--slaGaleq
#------------------------------------------------------------------------------
#  Transformation from ecliptic coordinates to J2000.0 equatorial coordinates.
sub slaEcleq {

   my ($dl, $db, $date) = @_;
   my ($dr, $dd); # return values

   if (!defined($date)) {
       $date = 51544.5;      # MJD of 2000/1/1
   }

   my (@rmat, @v1, @v2);

#  Spherical to Cartesian
   @v1 = slaDcs2c ( $dl, $db );

#  Ecliptic to equatorial
   @rmat =  slaEcmat ( $date );
   @v2 = slaDimxv ( \@rmat, \@v1 );

#  Mean of date to J2000
   @rmat = slaPrec ( 2000.0, slaEpj ( $date ) );
   @v1 = slaDimxv ( \@rmat, \@v2 );

#  Cartesian to spherical
   ( $dr, $dd ) = slaDcc2s ( @v1 );

#  Express in conventional ranges
   $dr = slaDranrm ( $dr );
   $dd = slaDrange ( $dd );

   return ( $dr, $dd );

} #eosub--slaEcleq
#------------------------------------------------------------------------------
#  Transformation from J2000.0 equatorial coordinates to ecliptic coordinates.
sub slaEqecl {

   my ($dr, $dd, $date) = @_;
   my ($dl, $db); # return values

   if (!defined($date)) {
       $date = 51544.5;      # MJD of 2000/1/1
   }

   my (@rmat, @v1, @v2);

#  Spherical to Cartesian
   @v1 = slaDcs2c ( $dr, $dd );

#  Mean J2000 to mean of date
   @rmat = slaPrec ( 2000.0, slaEpj ( $date ) );
   @v2 = slaDmxv ( \@rmat, \@v1 );

#  Equatorial to ecliptic
   @rmat = slaEcmat ( $date );
   @v1 = slaDmxv ( \@rmat, \@v2 );

#  Cartesian to spherical
   ( $dl, $db ) = slaDcc2s ( @v1 );

#  Express in conventional ranges
   $dl = slaDranrm ( $dl );
   $db = slaDrange ( $db );

   return ( $dl, $db );

} #eosub--slaEqecl
#------------------------------------------------------------------------------
#  Precess either FK4 or FK5 coordinates.
sub slaPreces {

   my ($sys, $ep0, $ep1, $ra, $dc) = @_;
   my (@pm, @v1, @v2);

#  Validate $sys
   if ($sys !~ /^FK[45]$/i) {
      return (-99, -99);
   } else {

   #  Generate appropriate precession matrix
      if (substr($sys, 2, 1) eq '4') {
         @pm = slaPrebn ( $ep0, $ep1 );
      } else {
         @pm = slaPrec ( $ep0, $ep1 );
      }

   #  Convert RA,Dec to x,y,z
      @v1 = slaDcs2c ( $ra, $dc );

   #  Precess
      @v2 = slaDmxv ( \@pm, \@v1 );

   #  Back to RA,Dec
      ( $ra, $dc ) = slaDcc2s ( @v2 );
      $ra = slaDranrm ( $ra );
      return ( $ra, $dc );
   }

} #eosub--slaPreces
#------------------------------------------------------------------------------
#  Besselian precession matrix.
sub slaPrebn {

   my ($bep0, $bep1) = @_;
   my ($bigt, $t, $tas2r, $w, $zeta, $z, $theta);

#  Interval between basic epoch B1850.0 and beginning epoch in TC
   $bigt  = ( $bep0 - 1850.0 ) / 100.0;

#  Interval over which precession required, in tropical centuries
   $t = ( $bep1 - $bep0 ) / 100.0;

#  Euler angles
   $tas2r = $t * $DAS2R;
   $w = 2303.5548 + ( 1.39720 + 0.000059 * $bigt ) * $bigt;
   $zeta = ($w + ( 0.30242 - 0.000269 * $bigt + 0.017996 * $t ) * $t ) * $tas2r;
   $z = ($w + ( 1.09478 + 0.000387 * $bigt + 0.018324 * $t ) * $t ) * $tas2r;
   $theta = ( 2005.1125 + ( -0.85294 - 0.000365 * $bigt ) * $bigt +
             ( -0.42647 - 0.000365 * $bigt - 0.041802 * $t ) * $t ) * $tas2r;

#  Rotation matrix
   return slaDeuler ( "ZYZ", -$zeta, $theta, -$z );

} #eosub--slaPreces
#------------------------------------------------------------------------------
#  Convert coordinates to unit vector.
sub slaDcs2c {

   my ($a, $b) = @_;
   my ($cosb);
   my ($v0, $v1, $v2);

   $cosb = cos ( $b );

   $v0 = cos ( $a ) * $cosb;
   $v1 = sin ( $a ) * $cosb;
   $v2 = sin ( $b );
   return ($v0, $v1, $v2);

} #eosub--slaDcs2c
#------------------------------------------------------------------------------
#  Convert unit vector to coordinates.
sub slaDcc2s {

   my ($x, $y, $z) = @_;
   my $r;
   my ($a, $b);

   $r = sqrt ( $x * $x + $y * $y );

   $a = ( $r != 0.0 ) ? atan2 ( $y, $x ) : 0.0;
   $b = ( $z != 0.0 ) ? atan2 ( $z, $r ) : 0.0;
   return ($a, $b);

} #eosub--slaDcc2s
#------------------------------------------------------------------------------
#  Multiply vector by rotation matrix.
sub slaDmxv {

    my ($dm, $va) = @_;
    my ($i, $j);
    my $w;
    my @vw;

   for ( $j = 0; $j < 3; $j++ ) {
      $w = 0.0;
      for ( $i = 0; $i < 3; $i++ ) {
         $w += $$dm[3 * $j + $i] * $$va[$i];
      }
      $vw[$j] = $w;
   }

   return @vw;

} #eosub--slaDmxv
#------------------------------------------------------------------------------
#  Multiply vector by inverse rotation matrix.
sub slaDimxv {

   my ($dm, $va) = @_;
   my ($i, $j, $w, @vw);

#  Inverse of matrix dm * vector va -> vector vw */
   for ( $j = 0; $j < 3; $j++ ) {
      $w = 0.0;
      for ( $i = 0; $i < 3; $i++ ) {
         $w += $$dm[3 * $i + $j] * $$va[$i];
      }
      $vw[$j] = $w;
   }

   return @vw;

} #eosub--slaDimxv
#------------------------------------------------------------------------------
#  Form the matrix of precession between two epochs (IAU 1976, FK5).
sub slaPrec {

   my ($ep0, $ep1) = @_;
   my ($t0, $t, $tas2r, $w, $zeta, $z, $theta);

#  Interval between basic epoch J2000.0 and beginning epoch (JC)
   $t0 = ( $ep0 - 2000.0 ) / 100.0;

#  Interval over which precession required (JC)
   $t =  ( $ep1 - $ep0 ) / 100.0;

#  Euler angles
   $tas2r = $t * $DAS2R;
   $w = 2306.2181 + ( ( 1.39656 - ( 0.000139 * $t0 ) ) * $t0 );
   $zeta = ($w + ( ( 0.30188 - 0.000344 * $t0 ) + 0.017998 * $t ) * $t )*$tas2r;
   $z = ($w + ( ( 1.09468 + 0.000066 * $t0 ) + 0.018203 * $t ) * $t ) * $tas2r;
   $theta = ( ( 2004.3109 + ( -0.85330 - 0.000217 * $t0 ) * $t0 )
          + ( ( -0.42665 - 0.000217 * $t0 ) - 0.041833 * $t ) * $t ) * $tas2r;

#  Rotation matrix
   return slaDeuler ( "ZYZ", -$zeta, $theta, -$z );

} #eosub--slaPrec
#------------------------------------------------------------------------------
#  Conversion of Modified Julian Date to Julian epoch.
sub slaEpj {

  return 2000.0 + ( $_[0] - 51544.5 ) / 365.25;

} #eosub--slaEpj
#------------------------------------------------------------------------------
#  Conversion of Modified Julian Date to Besselian epoch.
sub slaEpb {

  return 1900.0 + ( $_[0] - 15019.81352 ) / 365.242198781;

} #eosub--slaEpb
#------------------------------------------------------------------------------
#  Form the equatorial to ecliptic rotation matrix (IAU 1980 theory).
sub slaEcmat {

   my ($date) = @_;
   my ($t, $eps0);

#  Interval between basic epoch J2000.0 and current epoch (JC)
   $t = ( $date - 51544.5 ) / 36525.0;

#  Mean obliquity
   $eps0 = $DAS2R *
        ( 84381.448 + ( -46.8150 + ( -0.00059 + 0.001813 * $t ) * $t ) * $t );

#  Matrix
   return slaDeuler ( "X", $eps0, 0.0, 0.0 );

} #eosub--slaEcmat
#------------------------------------------------------------------------------
#  Form a rotation matrix from the Euler angles - three successive
#  rotations about specified Cartesian axes.
sub slaDeuler {

   my ($order, $phi, $theta, $psi) = @_;
   my ($j, $i, $l, $n, $k);
   my (@result, @rotn, $angle, $s, $c , $w, @wm);
   my ($axis);

#  Initialize result matrix
   for ( $j = 0; $j < 3; $j++ ) {
      for ( $i = 0; $i < 3; $i++ ) {
         $result[3*$i+$j] = ( $i == $j ) ? 1.0 : 0.0;
      }
   }

#  Establish length of axis string
   $l = length ( $order );

#  Look at each character of axis string until finished
   for ( $n = 0; $n < 3; $n++ ) {
      if ( $n <= $l ) {

         #  Initialize rotation matrix for the current rotation
         for ( $j = 0; $j < 3; $j++ ) {
            for ( $i = 0; $i < 3; $i++ ) {
               $rotn[3*$i+$j] = ( $i == $j ) ? 1.0 : 0.0;
            }
         }

         #  Pick up the appropriate Euler angle and take sine & cosine
         if ($n == 0) {
            $angle = $phi;
         } elsif ($n == 1) {
            $angle = $theta;
         } else {
            $angle = $psi;
         }
         $s = sin ( $angle );
         $c = cos ( $angle );

         #  Identify the axis
         $axis =  substr($order, $n, 1);

         if ( ( $axis eq 'X' ) || ( $axis eq 'x' ) || ( $axis eq '1' ) ) {

         #  Matrix for x-rotation
         #   Perl note: we don't do the index computation so that
         #              the user understands what's happening in the
         #              rotation matrix
            $rotn[3*1+1] = $c;
            $rotn[3*1+2] = $s;
            $rotn[3*2+1] = -$s;
            $rotn[3*2+2] = $c;
         }
         elsif ( ( $axis eq 'Y' ) || ( $axis eq 'y' ) || ( $axis eq '2' ) ) {

         #  Matrix for y-rotation
            $rotn[3*0+0] = $c;
            $rotn[3*0+2] = -$s;
            $rotn[3*2+0] = $s;
            $rotn[3*2+2] = $c;
         }
         elsif ( ( $axis eq 'Z' ) || ( $axis eq 'z' ) || ( $axis eq '3' ) ) {

         #  Matrix for z-rotation
            $rotn[3*0+0] = $c;
            $rotn[3*0+1] = $s;
            $rotn[3*1+0] = -$s;
            $rotn[3*1+1] = $c;
         } else {

         #  Unrecognized character - fake end of string
            $l = 0;
         }

         #  Apply the current rotation (matrix rotn x matrix result)
         for ( $i = 0; $i < 3; $i++ ) {
            for ( $j = 0; $j < 3; $j++ ) {
               $w = 0.0;
               for ( $k = 0; $k < 3; $k++ ) {
                  $w += $rotn[3*$i+$k] * $result[3*$k+$j];
               }
               $wm[3*$i+$j] = $w;
            }
         }
         for ( $j = 0; $j < 3; $j++ ) {
            for ( $i= 0; $i < 3; $i++ ) {
               $result[3*$i+$j] = $wm[3*$i+$j];
            }
         }
      }
   }

   return @result;

} #eosub--slaDeuler
#------------------------------------------------------------------------------
#  Convert from B1950 to J2000 for a given observational epoch.
sub slaFk45z {

   my ($r1950, $d1950, $bepoch) = @_;
   my ($r2000, $d2000); # return values

   if (!defined($bepoch)) {
      $bepoch = 1950.0;
   }

   my ($w, $i, $j);

#  Position and position+velocity vectors
   my (@r0, @a1, @v1, @v2);

#  Radians per year to arcsec per century
   my($pmf) = 100.0 * 60.0 * 60.0 * 360.0 / $D2PI;

#  Canonical constants:
#  vectors a and adot, and matrix m (only half of which is needed here)
   my (@a)  = ( -1.62557e-6,  -0.31919e-6, -0.13843e-6 );
   my (@ad) = (  1.245e-3,    -1.580e-3,   -0.659e-3 );
   my (@em) =
   (
       0.9999256782, -0.0111820611, -0.0048579477 ,
       0.0111820610,  0.9999374784, -0.0000271765 ,
       0.0048579479, -0.0000271474,  0.9999881997 ,
      -0.000551,     -0.238565,      0.435739     ,
       0.238514,     -0.002667,     -0.008541     ,
      -0.435623,      0.012254,      0.002117
   );

#  Spherical to Cartesian
   @r0 = slaDcs2c ( $r1950, $d1950 );

#  Adjust vector a to give zero proper motion in FK5
   $w = ( $bepoch - 1950.0 ) / $pmf;
   for ( $i = 0; $i < 3; $i++ ) {
      $a1[$i] = $a[$i] + $w * $ad[$i];
   }

#  Remove e-terms
   $w = $r0[0] * $a1[0] + $r0[1] * $a1[1] + $r0[2] * $a1[2];
   for ( $i = 0; $i < 3; $i++ ) {
      $v1[$i] = $r0[$i] - $a1[$i] + $w * $r0[$i];
   }

#  Convert position vector to Fricke system
   for ( $i = 0; $i < 6; $i++ ) {
      $w = 0.0;
      for ( $j = 0; $j < 3; $j++ ) {
         $w += $em[3 * $i + $j] * $v1[$j];
      }
      $v2[$i] = $w;
   }

#  Allow for fictitious proper motion in FK4
   $w = ( slaEpj ( slaEpb2d ( $bepoch ) ) - 2000.0 ) / $pmf;
   for ( $i = 0; $i < 3; $i++ ) {
      $v2[$i] += $w * $v2[$i + 3];
   }

#  Revert to spherical coordinates
   ( $r2000, $d2000 ) = slaDcc2s ( @v2 );
   $r2000 = slaDranrm ( $r2000 );
   return ( $r2000, $d2000 );

} #eosub--slaFk45z
#------------------------------------------------------------------------------
#  Convert J2000 to B1950 at a given Besselian epoch.  Note that
#  a proper motion is also returned.
sub slaFk54z {

    my ($r2000, $d2000, $bepoch) = @_;
    my ($r1950, $d1950, $dr1950, $dd1950);
    my ($zero) = 0.0;
    my ($r, $d, $px, $pv);

    if (!defined($bepoch)) {
       $bepoch = 1950.0;
    }

#  FK5 equinox J2000 (any epoch) to FK4 equinox B1950 epoch B1950
   ( $r, $d, $dr1950, $dd1950, $px, $pv ) =
     slaFk524 ( $r2000, $d2000, $zero, $zero, $zero, $zero );

#  Fictitious proper motion to epoch bepoch
   ( $r1950, $d1950 ) =
      slaPm ( $r, $d, $dr1950, $dd1950, $zero, $zero, 1950.0, $bepoch );
   return ( $r1950, $d1950, $dr1950, $dd1950 );

} #eosub--slaFk54z
#------------------------------------------------------------------------------
#  Convert J2000.0 FK5 star data to B1950.0 FK4.
sub slaFk524 {

    my ($r2000, $d2000, $dr2000, $dd2000, $p2000, $v2000) = @_;

#  Miscellaneous
   my ($r, $d, $ur, $ud, $px, $rv);
   my ($sr, $cr, $sd, $cd, $x, $y, $z, $w);
   my (@v1, @v2);
   my ($xd, $yd, $zd);
   my ($rxyz, $wd, $rxysq, $rxy);
   my ($i,$j);

#  Radians per year to arcsec per century
   my ($pmf) = 100.0 * 60.0 * 60.0 * 360.0 / $D2PI;

# Small number to avoid arithmetic problems
   my ($tiny) = 1.0e-30;

#  Canonical constants
#  km per sec to AU per tropical century
#    = 86400 * 36524.2198782 / 1.49597870e8
   my ($vf) = 21.095;

#  Constant vector and matrix (by rows)
   my (@a) = ( -1.62557e-6,   -0.31919e-6, -0.13843e-6,
                1.245e-3,     -1.580e-3,   -0.659e-3 );

   my (@emi) =
   (
        0.9999256795,              # emi[0][0]
        0.0111814828,              # emi[0][1]
        0.0048590039,              # emi[0][2]
       -0.00000242389840,          # emi[0][3]
       -0.00000002710544,          # emi[0][4]
       -0.00000001177742,          # emi[0][5]

       -0.0111814828,              # emi[1][0]
        0.9999374849,              # emi[1][1]
       -0.0000271771,              # emi[1][2]
        0.00000002710544,          # emi[1][3]
       -0.00000242392702,          # emi[1][4]
        0.00000000006585,          # emi[1][5]

       -0.0048590040,              # emi[2][0]
       -0.0000271557,              # emi[2][1]
        0.9999881946,              # emi[2][2]
        0.00000001177742,          # emi[2][3]
        0.00000000006585,          # emi[2][4]
       -0.00000242404995,          # emi[2][5]

       -0.000551,                  # emi[3][0]
        0.238509,                  # emi[3][1]
       -0.435614,                  # emi[3][2]
        0.99990432,                # emi[3][3]
        0.01118145,                # emi[3][4]
        0.00485852,                # emi[3][5]

       -0.238560,                  # emi[4][0]
       -0.002667,                  # emi[4][1]
        0.012254,                  # emi[4][2]
       -0.01118145,                # emi[4][3]
        0.99991613,                # emi[4][4]
       -0.00002717,                # emi[4][5]

        0.435730,                  # emi[5][0]
       -0.008541,                  # emi[5][1]
        0.002117,                  # emi[5][2]
       -0.00485852,                # emi[5][3]
       -0.00002716,                # emi[5][4]
        0.99996684                 # emi[5][5]
   );

#  Pick up J2000 data (units radians and arcsec/JC)
   $r = $r2000;
   $d = $d2000;
   $ur = $dr2000 * $pmf;
   $ud = $dd2000 * $pmf;
   $px = $p2000;
   $rv = $v2000;

#  Spherical to Cartesian
   $sr = sin ( $r );
   $cr = cos ( $r );
   $sd = sin ( $d );
   $cd = cos ( $d );

   $x = $cr * $cd;
   $y = $sr * $cd;
   $z = $sd;

   $w = $vf * $rv * $px;

   $v1[0] = $x;
   $v1[1] = $y;
   $v1[2] = $z;

   $v1[3] =  - $ur * $y - $cr * $sd * $ud + $w * $x;
   $v1[4] = $ur * $x - $sr * $sd * $ud + $w * $y;
   $v1[5] = $cd * $ud + $w * $z;

#  Convert position+velocity vector to BN system
   for ( $i = 0; $i < 6; $i++ ) {
      $w = 0.0;
      for ( $j = 0; $j < 6; $j++ ) {
         $w += $emi[6*$i+$j] * $v1[$j];
      }
      $v2[$i] = $w;
   }

#  Position vector components and magnitude
   $x = $v2[0];
   $y = $v2[1];
   $z = $v2[2];
   $rxyz = sqrt ( $x * $x + $y * $y + $z * $z );

#  Include e-terms
   $w = $x * $a[0] + $y * $a[1] + $z * $a[2];
   $x += $a[0] * $rxyz - $w * $x;
   $y += $a[1] * $rxyz - $w * $y;
   $z += $a[2] * $rxyz - $w * $z;

#  Recompute magnitude
   $rxyz = sqrt ( $x * $x + $y * $y + $z * $z );

#  Apply E-terms to both position and velocity
   $x = $v2[0];
   $y = $v2[1];
   $z = $v2[2];
   $w = $x * $a[0] + $y * $a[1] + $z * $a[2];
   $wd = $x * $a[3] + $y * $a[4] + $z * $a[5];
   $x += $a[0] * $rxyz - $w * $x;
   $y += $a[1] * $rxyz - $w * $y;
   $z += $a[2] * $rxyz - $w * $z;
   $xd = $v2[3] + $a[3] * $rxyz - $wd * $x;
   $yd = $v2[4] + $a[4] * $rxyz - $wd * $y;
   $zd = $v2[5] + $a[5] * $rxyz - $wd * $z;

#  Convert to spherical
   $rxysq = $x * $x + $y * $y;
   $rxy = sqrt ( $rxysq );

   if ( ( $x == 0.0 ) && ( $y == 0.0 ) ) {
      $r = 0.0;
   } else {
      $r = atan2 ( $y, $x );
      if ( $r < 0.0 ) {
          $r += $D2PI;
      }
   }
   $d = atan2 ( $z, $rxy );

   if ($rxy > $tiny) {
      $ur = ( $x * $yd - $y * $xd ) / $rxysq;
      $ud = ( $zd * $rxysq - $z * ( $x * $xd + $y * $yd ) ) /
           ( ( $rxysq + $z * $z ) * $rxy );
   }

#  Radial velocity and parallax
   if ( $px > $tiny ) {
      $rv = ( $x * $xd + $y * $yd + $z * $zd ) / ( $px * $vf * $rxyz );
      $px = $px / $rxyz;
   }

#  Return results
   return ($r, $d, $ur / $pmf, $ud / $pmf, $rv, $px);

} #eosub--slaFk524
#------------------------------------------------------------------------------
#  Apply corrections for proper motion to a star RA,Dec.
sub slaPm {

   my ($r0, $d0, $pr, $pd, $px, $rv, $ep0, $ep1) = @_;
   my ($r1, $d1 ); # return values

#  Km/s to AU/year multiplied by arc seconds to radians
   my ($vfr) = 0.21094502 * $DAS2R;

   my ($i, $w, @em, $t, @p);

#  Spherical to Cartesian
   @p = slaDcs2c ( $r0, $d0 );

#  Space motion (radians per year)
   $w = $vfr * $rv * $px;
   $em[0] = - $pr * $p[1] - $pd * cos ( $r0 ) * sin ( $d0 ) + $w * $p[0];
   $em[1] =   $pr * $p[0] - $pd * sin ( $r0 ) * sin ( $d0 ) + $w * $p[1];
   $em[2] =                 $pd * cos ( $d0 )               + $w * $p[2];

#  Apply the motion
   $t = $ep1 - $ep0;
   for ( $i = 0; $i < 3; $i++ ) {
      $p[$i] = $p[$i] + ($t * $em[$i]);
   }

#  Cartesian to spherical
   ( $r1, $d1 ) = slaDcc2s ( @p );
   $r1 = slaDranrm ( $r1 );
   return ( $r1, $d1 );

} #eosub--slaFk524
#------------------------------------------------------------------------------
#  Conversion of Julian epoch to Modified Julian Date.
sub slaEpj2d {

   return 51544.5 + ( $_[0] - 2000.0 ) * 365.25;

} #eosub--slaEpj2d
#------------------------------------------------------------------------------
#  Conversion of Besselian epoch to Modified Julian Date.
sub slaEpb2d {

   return 15019.81352 + ( $_[0] - 1900.0 ) * 365.242198781;

} #eosub--slaEpb2d
#------------------------------------------------------------------------------
#  Modulo function.
sub dmod {

   my($a,$b) = @_;
   return ( $a * $b > 0.0 ? $a - $b * int($a / $b)
                          : $a + $b * int(-$a / $b) );

} #eosub--dmod
#------------------------------------------------------------------------------
#  dsign(A,B) - magnitude of A with sign of B.
sub dsign {

   my($a,$b) = @_;
   return ( $b < 0.0 ) ? -$a : $a;

} #eosub--dsign
#------------------------------------------------------------------------------
#  Normalize angle into range 0-2 pi.
sub slaDranrm {

   my($angle) = @_;
   my($w) = dmod( $angle, $D2PI );
   return ( $w >= 0.0 ) ? $w : $w + $D2PI;

} #eosub--slaDranrm
#------------------------------------------------------------------------------
#  Normalize angle into range +/- pi.
sub slaDrange {

   my($angle) = @_;
   my($w) = dmod( $angle, $D2PI );
   return ( abs ( $w ) < $DPI ) ? $w : $w - dsign ( $D2PI, $angle );

} #eosub--slaDrange
#------------------------------------------------------------------------------
#  Scalar product of two 3-vectors.
sub slaDvdv {

   my ($va, $vb) = @_;
   return $va->[0] * $vb->[0] + $va->[1] * $vb->[1] + $va->[2] * $vb->[2];

} #eosub--slaDvdv
#------------------------------------------------------------------------------
#  Compute the e-terms (elliptic component of annual aberration) vector.
sub slaEtrms {

   my($ep) = @_;
   my(@ev); # return values

   my($t, $e, $e0, $p, $ek, $cp);

#  Julian centuries since B1950
   $t = ( $ep - 1950.0 ) * 1.00002135903e-2;

#  Eccentricity
   $e = 0.01673011 - ( 0.00004193 + 0.000000126 * $t ) * $t;

#  Mean obliquity
   $e0 = ( 84404.836 -
              ( 46.8495 + ( 0.00319 + 0.00181 * $t ) * $t ) * $t ) * $DAS2R;

#  Mean longitude of perihelion
   $p = ( 1015489.951 +
              ( 6190.67 + ( 1.65 + 0.012 * $t ) * $t ) * $t ) * $DAS2R;

#  E-terms
   $ek = $e * 20.49552 * $DAS2R;
   $cp = cos ( $p );
   $ev[0] = $ek * sin ( $p );
   $ev[1] = -$ek * $cp * cos ( $e0 );
   $ev[2] = -$ek * $cp * sin ( $e0 );

   return @ev;

} #eosub--slaEtrms
#------------------------------------------------------------------------------
#  Add the e-terms (elliptic component of annual aberration) to a
#  pre IAU 1976 mean place to conform to the old catalogue convention.
sub slaAddet {

   my($rm,$dm,$eq) = @_;
   my($rc,$dc); # return values

   my @a;    # Elliptic components of annual aberration vector
   my @v;    # Cartesian equivalant of RA,Dec
   my $i;

#  E-terms vector
   @a = slaEtrms ( $eq );

#  Spherical to Cartesian
   @v = slaDcs2c ( $rm, $dm );

#  Include the e-terms
   for ( $i = 0; $i < 3; $i++ ) {
       $v[$i] += $a[$i];
   }

#  Cartesian to spherical
   ($rc,$dc) = slaDcc2s ( @v );

#  Bring RA into conventional range
   $rc = slaDranrm ( $rc );

   return ( $rc, $dc );

} #eosub--slaAddet
#------------------------------------------------------------------------------
#  Remove the e-terms (elliptic component of annual aberration)
#  from a pre IAU 1976 catalogue RA,Dec to give a mean place.
sub slaSubet {

   my ($rc, $dc, $eq) = @_;
   my ($rm, $dm); # return values

   my (@a, @v, $f, $i);

#  E-terms
   @a = slaEtrms ( $eq );

#  Spherical to Cartesian
   @v = slaDcs2c ( $rc, $dc );

#  Include the e-terms
   $f = 1.0 + slaDvdv ( \@v, \@a );
   for ( $i = 0; $i < 3; $i++ ) {
      $v[$i] = $f * $v[$i] - $a[$i];
   }

#  Cartesian to spherical
   ( $rm, $dm ) = slaDcc2s ( @v );

#  Bring RA into conventional range
   $rm = slaDranrm ( $rm );

   return ( $rm, $dm );

} #eosub--slaSubet
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
